from collections import deque


class SpanningMixin(object):
    """Provides methods for checking whether an entity
    which has a defined start and end point over a single
    dimension contains or overlaps with another entity in
    that same dimension.
    """
    def __contains__(self, i):
        """Tests for point inclusion, `start <= i <= end`

        Parameters
        ----------
        i : Number
            The point to be tested

        Returns
        -------
        bool
        """
        return self.contains(i)

    def contains(self, i):
        """Tests for point inclusion, `start <= i <= end`

        Parameters
        ----------
        i : Number
            The point to be tested

        Returns
        -------
        bool
        """
        return self.start <= i <= self.end

    def overlaps(self, interval):
        """Tests whether another spanning entity with
        a defined start and end point overlaps with
        this spanning entity

        Parameters
        ----------
        interval : SpanningMixin

        Returns
        -------
        bool
        """
        cond = ((self.start <= interval.start and self.end >= interval.start) or (
            self.start >= interval.start and self.end <= interval.end) or (
            self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (
            self.start <= interval.start and self.end >= interval.start) or (
            self.start <= interval.end and self.end >= interval.end))
        return cond

    def overlap_size(self, interval):
        if interval.start in self and interval.end in self:
            return interval.end - interval.start
        elif self.start in interval and self.end in interval:
            return self.end - self.start
        if self.end >= interval.start:
            return self.end - interval.start
        elif self.start >= interval.end:
            return self.start - interval.end

    def is_contained_in_interval(self, interval):
        """Tests whether this spanning entity is
        completely contained inside the other entity

        Parameters
        ----------
        interval : SpanningMixin

        Returns
        -------
        bool
        """
        return self.start >= interval.start and self.end <= interval.end

    def contains_interval(self, interval):
        """Tests whether the other spanning entity is
        completely contained inside this entity

        Parameters
        ----------
        interval : SpanningMixin

        Returns
        -------
        bool
        """
        return self.start <= interval.start and self.end >= interval.end


class Interval(SpanningMixin):
    """A generic wrapper around data associated
    with a particular interval over a single dimension

    Attributes
    ----------
    data : dict
        A holder of arbitrary extra data not associated
        with any single member contained in this interval
    end : Number
        The end point of the interval described
    members : list
        A list of arbitrary objects which are associated with this
        interval.
    start : Number
        The start point of the interval described
    """
    def __init__(self, start, end, members=None, **kwargs):
        if members is None:
            members = []
        self.start = start
        self.end = end
        self.members = members
        self.data = kwargs

    def __repr__(self):
        return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)

    def __getitem__(self, k):
        return self.members[k]

    def __len__(self):
        return len(self.members)

    def __iter__(self):
        return iter(self.members)


class IntervalTreeNode(object):
    """A single node in an Interval Tree. The
    root node of an interval tree can be treated
    as the tree itself.

    Attributes
    ----------
    center : Number
        The center point of this node's collection
    contained : list
        The list of Interval-like objects which
        were aggregated under this node. This collection
        of intervals all span this node's center.
    end : Number
        The end point of this node's collection or it's
        rightmost child's end point, whichever is larger
    left : IntervalTreeNode
        The left child node of this node. May be `None`
        if this node is a leaf node. Contains all
        intervals whose end point is less than this
        node's center
    level : int
        Depth in the tree
    parent : IntervalTree
        This node's parent node. May be None if this
        node is the root node.
    right : IntervalTree
        The right child node of this node. May be `None`
        if this node is a leaf node. Contains all
        intervals whose start point is greater than this
        node's center
    start : Number
        The start point of this node's collection or it's
        leftmost child's start point, whichever is smaller
    """
    def __init__(self, center, left, contained, right, level=0, parent=None):
        self.center = center
        self.left = left
        self.contained = contained
        self.right = right
        self.level = level
        self.parent = parent

        start = float('inf')
        end = -float('inf')
        i = 0

        for interval in self.contained:
            i += 1
            if interval.start < start:
                start = interval.start
            if interval.end > end:
                end = interval.end

        if i > 0:
            self.start = start
            self.end = end
        else:
            self.start = self.end = center

    def contains_point(self, x):
        """Returns the list of contained intervals for all
        nodes which contain the point `x`.

        Parameters
        ----------
        x : Number
            The query point

        Returns
        -------
        list
            A list of objects which span `x` from all spanning
            nodes' `contained` list.
        """
        inner = [
            i for i in self.contained
            if i.start <= x <= i.end
        ]
        if self.left is not None and self.left.start <= x <= self.left.end:
            return self.left.contains_point(x) + inner
        if self.right is not None and self.right.start <= x <= self.right.end:
            return self.right.contains_point(x) + inner
        else:
            return inner

    def _overlaps_interval(self, start, end):
        result = []
        query = Interval(start, end)
        for i in self.contained:
            if query.overlaps(i):
                result.append(i)
        return result

    def overlaps(self, start, end):
        """Returns the list of all contained intervals which
        overlap with the interval described by `start` and `end`

        Parameters
        ----------
        start : Number
            The start of the query interval
        end : Number
            The end of the query interval

        Returns
        -------
        list
            A list of all objects which overlap the argument interval
            from all spanning nodes' `contained` list.
        """
        result = []
        if start > self.end:
            return result
        if end < self.start:
            return result
        elif start <= self.start:
            if end < self.start:
                return result
            else:
                if self.left is not None:
                    result.extend(self.left.overlaps(start, end))
                result.extend(self._overlaps_interval(start, end))
                if self.right is not None and end >= self.right.start:
                    result.extend(self.right.overlaps(start, end))
        elif start > self.start:
            if self.left is not None and self.left.end >= start:
                result.extend(self.left.overlaps(start, end))
            result.extend(self._overlaps_interval(start, end))
            if self.right is not None and end >= self.right.start:
                result.extend(self.right.overlaps(start, end))
        elif end > self.start:
            if self.left is not None:
                result.extend(self.left.overlaps(start, end))
            result.extend(self._overlaps_interval(start, end))
            if self.right is not None and end >= self.right.start:
                result.extend(self.right.overlaps(start, end))
        return result

    def node_contains(self, point):
        return self.start <= point <= self.end

    def __repr__(self):
        return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (
            self.level, self.center, self.start, self.end)

    def __eq__(self, other):
        if other is None:
            return False
        else:
            result = self.contained == other.contained
            if result:
                result = self.left == other.left
                if result:
                    result = self.right == other.right
            return result

    def __hash__(self):
        return hash((self.start, self.center, self.right, self.level))

    def __diagnostic_eq__(self, other):  # pragma: no cover
        if other is None:
            return False
        else:
            result = self.contained == other.contained
            if result:
                try:
                    result = self.left.__diagnostic_eq__(other.left)
                except AttributeError:
                    result = self.left == other.left
                if result:
                    try:
                        result = self.right.__diagnostic_eq__(other.right)
                    except AttributeError:
                        result = self.right == other.right
                    if not result:
                        print(self.right, "r!=", other.right)
                else:
                    print(self.left, "l!=", other.left)
            else:
                print(self, "!=", other)
            return result

    def _update_bounds(self):
        changed = False
        if self.left is not None and self.left.start < self.start:
            changed = True
            self.start = self.left.start
        if self.right is not None and self.right.end > self.end:
            changed = True
            self.end = self.right.end
        if self.parent is not None:
            self.parent._update_bounds()
        return changed

    def insert(self, interval):
        insert_in_self = False
        if self.node_contains(interval.start):
            if self.left is not None and self.left.node_contains(interval.end):
                return self.left.insert(interval)
            elif (self.right is not None and self.right.node_contains(interval.end) and
                  self.right.node_contains(interval.start)):
                return self.right.insert(interval)
            else:
                insert_in_self = True
        elif self.node_contains(interval.end):
            if self.right is not None and self.right.node_contains(interval.start):
                return self.right.insert(interval)
            else:
                insert_in_self = True
        if not insert_in_self and self.parent is None:
            insert_in_self = True
        if insert_in_self:
            self.contained.append(interval)
            changed = False
            if interval.start < self.start:
                self.start = interval.start
                changed = True
            if interval.end > self.end:
                self.end = interval.end
                changed = True
            if changed:
                self._update_bounds()

    def flatten(self):
        # perform an infix traversal of the tree and collect
        # the elements
        items = []
        stack = deque([])
        current = self
        done = False
        while not done:
            if current is not None:
                stack.append(current)
                current = current.left
            else:
                if stack:
                    current = stack.pop()
                    items.extend(current.contained)
                    current = current.right
                else:
                    done = True
        return items

    def balance(self):
        items = self.flatten()
        tree = self.build(items)
        self.left = tree.left
        self.right = tree.right
        self.contained = tree.contained
        self.start = tree.start
        self.end = tree.end
        self.center = tree.center
        self.left.parent = self
        self.right.parent = self
        self._update_bounds()


def iterative_build_interval_tree(cls, intervals):
    """Builds an IntervalTreeNode hierarchy from `intervals`. This
    iterative method is preferable to avoid recursion limits.

    Parameters
    ----------
    intervals : Iterable of Intervals
        The set of Interval-like objects to use to construct
        the Interval Tree

    Returns
    -------
    IntervalTreeNode
        The root of the constructed Interval Tree
    """
    stack = []
    root = cls(0, None, [], None, -1)
    if not intervals:
        return root

    stack.append((root, intervals, "left"))

    while stack:
        parent, members, side = stack.pop()
        if len(members) > 0:
            centers = [
                float(i.start + i.end) / 2. for i in members
            ]
            left = []
            right = []
            contained = []
            center = sum(centers) / (len(centers))
            if len(members) > 20:
                for i in members:
                    if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
                        contained.append(i)
                    elif center > i.end:
                        left.append(i)
                    elif center < i.start:
                        right.append(i)
                    else:
                        contained.append(i)
            else:
                contained = members[:]

            if len(right) == len(members) or len(left) == len(members):
                contained = members[:]
                left = []
                right = []
            node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)
            if side == 'left':
                parent.left = node
                parent.start = min(node.start, parent.start)
                up = parent.parent
                while up is not None:
                    if up.start > node.start:
                        up.start = node.start
                    up = up.parent
            elif side == 'right':
                parent.right = node
                parent.end = max(node.end, parent.end)
                up = parent.parent
                while up is not None:
                    if up.end < node.end:
                        up.end = node.end
                    up = up.parent
            else:
                raise ValueError(side)
            stack.append((node, left, "left"))
            stack.append((node, right, "right"))
    return root.left


IntervalTreeNode.build = classmethod(iterative_build_interval_tree)


def recursive_build_interval_tree(cls, intervals, level=0):  # pragma: no cover
    if len(intervals) > 0:
        centers = [
            float(i.start + i.end) / 2. for i in intervals
        ]
        left = []
        right = []
        contained = []
        center = sum(centers) / (len(centers) + 1.)
        if len(intervals) > 20:
            for i in intervals:
                if center >= i.end:
                    left.append(i)
                elif center < i.start:
                    right.append(i)
                else:
                    contained.append(i)
            left = cls.recursive_build_interval_tree(left, level + 1)
            right = cls.recursive_build_interval_tree(right, level + 1)
        else:
            contained = intervals
            left = None
            right = None

        inst = cls(center, left, contained, right, level)
        if left is not None:
            inst.left.parent = inst
            inst.start = min(inst.start, left.start)
        if right is not None:
            inst.right.parent = inst
            inst.end = (max(inst.end, right.end))
        return inst
    else:
        return None


IntervalTreeNode.recursive_build_interval_tree = classmethod(
    recursive_build_interval_tree)




cdef class Configuration(object):
    cdef:
        public str N_TERM_DEFAULT
        public str C_TERM_DEFAULT
        public int FRAG_OFFSET
        public bint ALLOW_MODIFIED_ASPARAGINE
        public bint PARTIAL_HEXNAC_LOSS
        public bint EXCLUDE_B1


from typing import Optional
from glycopeptidepy.utils.memoize import memoize

from .base import MoleculeBase
from .modification import Modification
from .composition import Composition, formula

from six import string_types as basestring


class TerminalGroup(MoleculeBase):
    __slots__ = ("base_composition", "mass", "_modification")

    base_composition: Composition
    mass: float
    _modification: Optional[Modification]

    def __init__(self, base_composition, modification=None):
        if not isinstance(base_composition, Composition):
            base_composition = Composition(base_composition)
        self.mass = 0
        self.base_composition = base_composition
        self._modification = None
        if modification is not None:
            self.modification = modification
        self.mass = self._calculate_mass()

    def _calculate_mass(self) -> float:
        base_mass = self.base_composition.mass
        mod = self.modification
        if mod is not None:
            base_mass += mod.mass
        return base_mass

    def clone(self):
        return self.__class__(self.base_composition, self.modification)

    def __reduce__(self):
        return self.__class__, (self.base_composition, self.modification)

    @property
    def modification(self):
        return self._modification

    @modification.setter
    def modification(self, value):
        if value is not None:
            new_mass = value.mass
        else:
            new_mass = 0
        if self._modification is not None:
            old_mass = self._modification.mass
        else:
            old_mass = 0
        self.mass += new_mass - old_mass
        self._modification = value

    def modify(self, modification):
        return self.__class__(self.base_composition, modification)

    @property
    def composition(self):
        modification = self.modification
        if modification is None:
            return self.base_composition
        mod_comp = modification.composition
        return self.base_composition + mod_comp

    def __repr__(self):
        template = "{self.__class__.__name__}({self.base_composition}, {self.modification})"
        return template.format(self=self)

    def __str__(self):
        if self.modification is not None:
            return str(self.modification)
        return formula(self.base_composition)

    def __eq__(self, other):
        if other is None:
            return False
        try:
            return (self.base_composition == other.base_composition) and (self.modification == other.modification)
        except AttributeError:
            if isinstance(other, basestring):
                return self.composition == Modification(other).composition
            else:
                try:
                    return self.composition == other.composition
                except AttributeError:
                    return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(formula(self.composition))

    def serialize(self):
        return str(self)


try:
    from glycopeptidepy._c.structure.base import TerminalGroup
except ImportError:
    pass


@memoize(100)
def _make_terminal_group(base_composition_formula, modification=None):
    return TerminalGroup(base_composition_formula, modification)



import unittest
import pickle

import glypy

from glycopeptidepy.structure import sequence, fragment


class TestPeptideFragment(unittest.TestCase):
    def test_fragment_hashing(self):
        seq = sequence.PeptideSequence("LVPVPITN(N-Glycosylation)ATLDQITGK{Hex:5; HexNAc:4; Neu5Ac:2}")
        frag_map = dict()
        for position in seq.get_fragments('b'):
            for frag in position:
                frag_map[frag] = 1
        seq = sequence.PeptideSequence("LVPVPITN(N-Glycosylation)ATLDQITGK{Hex:5; HexNAc:4; Neu5Ac:2}")
        for position in seq.get_fragments('b'):
            for frag in position:
                self.assertEqual(frag_map[frag], 1)
                if "HexNAc" in frag.name:
                    assert frag.glycosylation_size > 0
                    assert frag.is_glycosylated
                else:
                    assert not frag.is_glycosylated
                    assert frag.name == frag.base_name()

    def test_pickle(self):
        seq = sequence.PeptideSequence("LVPVPITN(N-Glycosylation)ATLDQITGK{Hex:5; HexNAc:4; Neu5Ac:2}")
        for position in seq.get_fragments('b'):
            for frag in position:
                assert frag == frag.clone()
                assert frag == pickle.loads(pickle.dumps(frag))


class TestStubFragment(unittest.TestCase):
    def test_pickle(self):
        seq = sequence.PeptideSequence("LVPVPITN(N-Glycosylation)ATLDQITGK{Hex:5; HexNAc:4; Neu5Ac:2}")
        for frag in seq.stub_fragments():
            assert frag == frag.clone()
            assert frag == pickle.loads(pickle.dumps(frag))

    def test_naming(self):
        gc = glypy.GlycanComposition(Hex=5, HexNAc=4, NeuAc=2)
        gc.composition_offset = glypy.Composition()
        peptide = sequence.PeptideSequence("LVPVPITNATLDQITGK")
        name = fragment.StubFragment.build_name_from_composition(gc)
        frag = fragment.StubFragment(
            name, peptide.mass + gc.mass(),
            fragment.IonSeries.stub_glycopeptide,
            peptide.total_composition() + gc.total_composition(),
            glycosylation=gc)
        assert frag.name == "peptide+Hex5HexNAc4Neu5Ac2"
        assert frag.base_name() == "peptide+Hex5HexNAc4Neu5Ac2"
        assert abs(frag.mass - (peptide.mass + gc.mass())) < 1e-3
        shift = fragment.ChemicalShift("Na1H-1", glypy.Composition("Na1H-1"))
        frag.chemical_shift = shift
        assert frag.name == "peptide+Hex5HexNAc4Neu5Ac2Na1H-1"
        assert frag.base_name() == "peptide+Hex5HexNAc4Neu5Ac2"
        assert abs(frag.mass - (peptide.mass + gc.mass() + shift.mass)) < 1e-3


if __name__ == '__main__':
    unittest.main()



import unittest

from glycopeptidepy.structure import sequence

from glycopeptidepy.test.sequence_test_suite import PeptideSequenceSuiteBase, TestSequenceParser


class TestPeptideSequence(PeptideSequenceSuiteBase, unittest.TestCase):
    def parse_sequence(self, seqstr):
        return sequence.PeptideSequence(seqstr)


class TestNamedSequence(PeptideSequenceSuiteBase, unittest.TestCase):
    def parse_sequence(self, seqstr):
        return sequence.NamedSequence("spam", seqstr)


class TestAnnotatedSequence(PeptideSequenceSuiteBase, unittest.TestCase):
    def parse_sequence(self, seqstr):
        return sequence.AnnotatedSequence("spam", seqstr, annotations={"color": "green"})


if __name__ == '__main__':
    unittest.main()



def simple_repr(self):  # pragma: no cover
    template = "{self.__class__.__name__}({d})"
    d = [
        "%s=%r" % (k, v) if v is not self else "(...)" for k, v in sorted(
            self.__dict__.items(), key=lambda x: x[0])
        if not k.startswith("_") and not callable(v)]
    return template.format(self=self, d=', '.join(d))



cimport cython

from glypy.composition.ccomposition cimport CComposition
from glycopeptidepy._c.structure.base cimport (
    PeptideSequenceBase,
    SequencePosition, TerminalGroup, AminoAcidResidueBase,
    ModificationBase)


cdef AminoAcidResidueBase _parse_residue(basestring residue_string)

cdef TerminalGroup _make_terminal_group(basestring base_composition_formula, ModificationBase modification=*)


cdef class _PeptideSequenceCore(PeptideSequenceBase):

    cpdef _init_from_parsed_string(self, list seq_list, glycan=*, n_term=*, c_term=*)

    cpdef _init_from_string(self, sequence, parser_function)

    cpdef _init_from_components(self, list seq_list, glycan_composition=*, ModificationBase n_term=*, ModificationBase c_term=*)

    cpdef _init_termini(self, TerminalGroup n_term, TerminalGroup c_term)

    cpdef _initialize_fields(self)

    cdef size_t get_size(self)
    cdef TerminalGroup get_n_term(self)
    cdef void set_n_term(self, _value)
    cdef TerminalGroup get_c_term(self)
    cdef void set_c_term(self, _value)

    cpdef basestring get_sequence(self, bint include_glycan=*, bint include_termini=*, str implicit_n_term=*, str implicit_c_term=*)
    cpdef clone(self)

    cpdef bint base_sequence_equality(self, _PeptideSequenceCore other)
    cpdef bint modified_sequence_equality(self, _PeptideSequenceCore other)
    cpdef bint full_structure_equality(self, _PeptideSequenceCore other)

    cpdef CComposition peptide_composition(self)
    cpdef CComposition total_composition(self)

    cdef double get_peptide_backbone_mass(self)
    cdef double get_total_mass(self)

cpdef add_modification(_PeptideSequenceCore self, position, modification_type)


import re


from importlib import resources
from typing import Optional, Tuple

from glypy import Composition

from glycopeptidepy.structure.residue import long_to_symbol
from glycopeptidepy.structure.modification import (
    ModificationRule,
    ModificationTarget,
    SequenceLocation,
    ModificationTable)


class UniProtPTMListParser(object):  # pragma: no cover
    def __init__(self, path=None):
        self.path = path
        self.handle = open(path)
        self._find_starting_point()

    def _find_starting_point(self):
        self.handle.seek(0)
        for line in self.handle:
            if line.startswith("___"):
                break

    def _next_line(self) -> Tuple[str, str]:
        line = next(self.handle)
        line = line.strip("\n")
        tokens = re.split(r"\s+", line, maxsplit=1)
        if len(tokens) == 1:
            return tokens[0], ""
        else:
            typecode, content = tokens
            return typecode, content

    def _formula_parser(self, formula: str) -> Composition:
        counts = dict()
        for symbol, count in re.findall(r"([A-Za-z]+)(-?\d+)", formula):
            count = int(count)
            counts[symbol] = count
        return Composition(counts)

    def _translate_position(self, position):
        t = {
            'Anywhere': SequenceLocation.anywhere,
            'N-terminal': SequenceLocation.n_term,
            'C-terminal': SequenceLocation.c_term
        }
        return t.get(position, 'Anywhere')

    def _translate_amino_acid(self, target):
        try:
            symbol = long_to_symbol[target]
        except KeyError:
            return None
        return symbol

    def parse_entry(self) -> Optional[ModificationRule]:
        ptm_id = None
        accession = None
        # feature_key = None
        mass_difference = None
        correction_formula = None
        target = None
        position = None

        keywords = []
        crossref = []
        while True:
            typecode, line = self._next_line()
            if typecode == "ID":
                ptm_id = line
            elif typecode == "AC":
                accession = line
            elif typecode == "FT":
                # feature_key = line
                pass
            elif typecode == 'TG':
                target = self._translate_amino_acid(line.strip("."))
            elif typecode == 'PP':
                position = self._translate_position(line.strip('.'))
            elif typecode == "CF":
                correction_formula = self._formula_parser(line)
            elif typecode == "MM":
                mass_difference = float(line)
            elif typecode == "KW":
                keywords.append(line.strip("."))
            elif typecode == "DR":
                alias = ':'.join(line.strip('.').split("; "))
                if alias.startswith("PSI-MOD:"):
                    alias = alias.replace("PSI-MOD:", "MOD:")
                crossref.append(alias)
            elif typecode == "//":
                break
        if mass_difference is None or correction_formula is None:
            return None
        mod_target = ModificationTarget([target] if target else None, position)
        rule = ModificationRule(
            [mod_target], ptm_id, monoisotopic_mass=mass_difference, composition=correction_formula,
            alt_names={accession} | set(crossref), categories=keywords)
        rule.preferred_name = rule.name = ptm_id
        return rule

    def build_table(self) -> ModificationTable:
        table = {}
        while True:
            try:
                entry = self.parse_entry()
                if entry is None:
                    continue
                table[entry.name] = entry
            except StopIteration:
                break
        return ModificationTable(table.values())


def load(path=None):
    if path is None:
        with resources.path("glycopeptidepy.io.cv.data", "uniprot_ptmlist.txt") as path:
            parser = UniProtPTMListParser(path)
            table = parser.build_table()
            return table
    else:
        parser = UniProtPTMListParser(path)
        table = parser.build_table()
        return table



import warnings
import re

from lxml import etree
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, backref, object_session
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy import (Numeric, Unicode,
                        Column, Integer, ForeignKey,
                        UnicodeText, Boolean, event)
from sqlalchemy import exc as sa_exc
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from glypy.composition.composition import Composition, _make_isotope_string

Base = declarative_base()

_unimod_xml_download_url = "http://www.unimod.org/xml/unimod_tables.xml"

try:
    basestring
except NameError:
    basestring = (str, bytes)


CompositionType = Composition


def simple_repr(self):  # pragma: no cover
    template = "{self.__class__.__name__}({d})"
    items = [(k, v) for k, v in self.__dict__.items() if not k.startswith("_")]
    parts = []
    for key, value in items:
        if key.endswith("_id"):
            try:
                value = getattr(self, key[:-3])
                key = key[:-3]
            except AttributeError:
                pass
        parts.append("%s=%r" % (key, value))

    return template.format(self=self, d=', '.join(parts))

Base.__repr__ = simple_repr


def attr_repr(attr_name):  # pragma: no cover
    def __repr__(self):
        return str(getattr(self, attr_name))

    def applier(cls):
        cls.__repr__ = __repr__
        return cls

    return applier


def remove_namespace(doc, namespace):
    """Remove namespace in the passed document in place."""
    ns = u'{%s}' % namespace
    nsl = len(ns)
    for elem in doc.getiterator():
        if elem.tag.startswith(ns):
            elem.tag = elem.tag[nsl:]


def preprocess_xml(doc_path):
    '''
    Parse and drop namespaces from an XML document.

    Parameters
    ----------
    doc_path: str

    Returns
    -------
    etree.ElementTree
    '''
    tree = etree.parse(doc_path)
    root = tree.getroot()
    for ns in root.nsmap.values():
        remove_namespace(tree, ns)
    return tree


def _formula_parser(formula, session):
    '''
    Parse a unimod formula composed of elements,
    isotopes, and other bricks.

    In order to look up a Brick's composition, this
    function must have access to a session.

    Parameters
    ----------
    formula: str
        A Unimod formula of the form `A(n) B(m)...`
        where A, B, ... are element names or bricks and
        (n), (m)... are parenthesized possibly signed integers or
        omitted in which case they are interpreted as 1
    session: Session
        An active SQLAlchemy session for looking up bricks in the database

    Returns
    -------
    CompositionType
    '''
    composition = CompositionType()
    for token in formula.split(" "):
        match = re.search(r"(?P<isotope>\d+)?(?P<elemet>[^\(]+)(?:\((?P<count>-?\d+)\))?", token)
        if match:
            isotope, element, count = match.groups()
            if count is not None:
                count = int(count)
            else:
                count = 1
            if isotope is not None:
                name = _make_isotope_string(element, int(isotope))
            else:
                name = element
            if name == '0':
                continue
            is_brick = session.query(Brick).filter(Brick.brick == name).first()
            if is_brick is None:
                composition[str(name)] += count
            else:
                composition += is_brick.composition * count
    return composition


def _composition_listener(attr):
    '''
    Attach event listeners to an InstrumentedAttribute
    to trigger formula parsing on load and on change.
    '''
    @event.listens_for(attr, "set")
    def _update_composition_from_formula(target, value, oldvalue, initiator):
        session = object_session(target)
        if value == "" or value is None:
            return
        # If the object hasn't been associated with a session,
        # we can't look up bricks.
        if session is None:
            return
        target.composition = _formula_parser(value, session)

    @event.listens_for(attr.class_, "load")
    def _update_composition_on_load(target, context):
        value = getattr(target, attr.prop.key)
        if value == "" or value is None:
            return
        session = object_session(target)
        target.composition = _formula_parser(value, session)


def has_composition(attr_name):
    '''
    A decorator to simplify flagging a Model with a column
    to be treated as a formula for parsing. Calls :func:`_composition_listener`
    internally.
    '''
    def decorator(model):
        _composition_listener(getattr(model, attr_name))
        return model
    return decorator


class HasFullNameMixin(object):
    '''
    A simple mixin to standardize equality operators
    for models with a :attr:`full_name` attribute.
    '''
    def __eq__(self, other):
        try:
            return self.full_name == other.full_name
        except AttributeError:
            return self.full_name == other

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.full_name)


class AlternativeName(Base):
    __tablename__ = "AlternativeName"

    _tag_name = "alt_names_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib["record_id"]),
            alt_name=attrib['alt_name'],
            modification_id=int(attrib['mod_key'])
            )
        return inst

    id = Column(Integer, primary_key=True)
    alt_name = Column(Unicode(256), index=True)
    modification_id = Column(Integer, ForeignKey("Modification.id"), index=True)


class AminoAcid(Base, HasFullNameMixin):
    __tablename__ = "AminoAcid"

    _tag_name = "amino_acids_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            full_name=attrib['full_name'],
            one_letter=attrib['one_letter'],
            three_letter=attrib['three_letter'],
            num_H=int(attrib["num_H"]),
            num_O=int(attrib["num_O"]),
            num_C=int(attrib["num_C"]),
            num_N=int(attrib["num_N"]),
            num_S=int(attrib["num_S"]),
            )
        return inst

    id = Column(Integer, primary_key=True)
    num_H = Column(Integer)
    num_O = Column(Integer)
    num_C = Column(Integer)
    num_N = Column(Integer)
    num_S = Column(Integer)
    full_name = Column(Unicode(25), index=True)
    one_letter = Column(Unicode(10), index=True)
    three_letter = Column(Unicode(10), index=True)


@attr_repr("classification")
class Classification(Base):
    __tablename__ = "Classification"

    _tag_name = "classifications_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            classification=attrib["classification"]
            )
        return inst

    id = Column(Integer, primary_key=True)
    classification = Column(Unicode(30), index=True)


@attr_repr("position")
class Position(Base):
    __tablename__ = "Position"

    _tag_name = "positions_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            position=attrib['position']
            )
        return inst

    def __eq__(self, other):
        try:
            return self.position == other.position
        except AttributeError:
            return self.position == other

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.position)

    id = Column(Integer, primary_key=True)
    position = Column(Unicode(20), index=True)


class Brick(Base, HasFullNameMixin):
    __tablename__ = "Brick"

    _tag_name = "bricks_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            brick=attrib['brick'],
            full_name=attrib['full_name']
            )
        return inst

    id = Column(Integer, primary_key=True)
    brick = Column(Unicode(64), index=True)
    full_name = Column(Unicode(128), index=True)

    elements = relationship("BrickToElement")

    @property
    def composition(self):
        composition = CompositionType()
        for element_relation in self.elements:
            symbol = element_relation.element
            isotope, element = re.search(r"(?P<isotope>\d*)?(?P<element>\S+)", symbol).groups()
            if isotope != "":
                isotope = int(isotope)
                iso_str = _make_isotope_string(element, isotope)
            else:
                iso_str = element
            count = element_relation.count
            composition[str(iso_str)] = count
        return composition


class Fragment(Base):
    __tablename__ = "Fragment"

    _tag_name = "fragments_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            modification_id=int(attrib["mod_key"])
            )
        return inst

    id = Column(Integer, primary_key=True)
    modification_id = Column(Integer, ForeignKey("Modification.id"), index=True)

    _fragment_composition = relationship("FragmentComposition")

    @property
    def composition(self):
        composition = CompositionType()
        session = object_session(self)
        for fragment_composition_relation in self._fragment_composition:
            symbol = fragment_composition_relation.brick_string
            isotope, element = re.search(r"(?P<isotope>\d*)?(?P<element>\S+)", symbol).groups()
            count = fragment_composition_relation.count
            if count is not None:
                count = int(count)
            else:
                count = 1
            if isotope != "":
                name = _make_isotope_string(element, isotope)
            else:
                name = element
            is_brick = session.query(Brick).filter(Brick.brick == name).first()
            if is_brick is None:
                composition[str(name)] += count
            else:
                composition += is_brick.composition * count
        return composition


class FragmentComposition(Base):
    __tablename__ = "FragmentComposition"

    _tag_name = "fragment_comp_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            brick_string=attrib["brick"],
            fragment_id=int(attrib["fragments_key"]),
            count=int(attrib["num_brick"])
            )
        return inst

    id = Column(Integer, primary_key=True)
    brick_string = Column(Unicode(64), ForeignKey(Brick.brick), index=True)
    fragment_id = Column(Integer, ForeignKey("Fragment.id"), index=True)
    count = Column(Integer)


class ModificationToBrick(Base):
    __tablename__ = "ModificationToBrick"

    _tag_name = "mod2brick_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            brick_string=(attrib['brick']),
            modification_id=int(attrib["mod_key"]),
            count=int(attrib["num_brick"])
            )
        return inst

    id = Column(Integer, primary_key=True)
    brick_string = Column(Unicode(64), ForeignKey(Brick.brick), index=True)
    modification_id = Column(Integer, ForeignKey("Modification.id"), index=True)
    count = Column(Integer)


class BrickToElement(Base):
    __tablename__ = "BrickToElement"

    _tag_name = "brick2element_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            brick_id=int(attrib['brick_key']),
            count=int(attrib["num_element"]),
            element=attrib['element']
            )
        return inst

    id = Column(Integer, primary_key=True)
    brick_id = Column(Integer, ForeignKey(Brick.id), index=True)
    element = Column(Unicode(16), ForeignKey("Element.element"), index=True)
    element_obj = relationship("Element", uselist=False)
    count = Column(Integer)


class Element(Base, HasFullNameMixin):
    __tablename__ = "Element"

    _tag_name = "elements_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            average_mass=float(attrib["avge_mass"]),
            monoisotopic_mass=float(attrib["mono_mass"]),
            full_name=attrib["full_name"],
            element=attrib["element"]

            )
        return inst

    id = Column(Integer, primary_key=True)
    average_mass = Column(Numeric(12, 6, asdecimal=False))
    monoisotopic_mass = Column(Numeric(12, 6, asdecimal=False))
    full_name = Column(Unicode(64), index=True)
    element = Column(Unicode(16), index=True)


@has_composition("_composition")
class Modification(Base, HasFullNameMixin):
    __tablename__ = "Modification"

    _tag_name = "modifications_row"

    id = Column(Integer, primary_key=True)
    username_of_poster = Column(Unicode(128))
    average_mass = Column(Numeric(12, 6, asdecimal=False), index=True)
    ex_code_name = Column(Unicode(64), index=True)
    monoisotopic_mass = Column(Numeric(12, 6, asdecimal=False), index=True)
    full_name = Column(Unicode(128), index=True)
    code_name = Column(Unicode(128), index=True)
    _composition = Column(Unicode(128), index=True)
    approved = Column(Boolean, index=True)

    notes = relationship("MiscNotesModifications")
    specificities = relationship("Specificity")
    bricks = relationship(ModificationToBrick)
    _fragments = relationship(Fragment)

    _alt_names = relationship(AlternativeName, backref=backref("modification"))
    # Maps the list of AlternativeName instances loaded dynamically from _alt_names
    # into a list of plain strings, since the AlternativeName type contains no
    # additional information.
    alternative_names = association_proxy("_alt_names", "alt_name")
    fragments = association_proxy("_fragments", "composition")

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib["record_id"]),
            username_of_poster=attrib["username_of_poster"],
            average_mass=float(attrib["avge_mass"]),
            monoisotopic_mass=float(attrib["mono_mass"]),
            ex_code_name=attrib["ex_code_name"],
            code_name=attrib["code_name"],
            full_name=attrib["full_name"],
            approved=bool(int(attrib['approved'])),
            _composition=attrib["composition"]
            )
        for note in tag:
            if note.tag == MiscNotesModifications._tag_name:
                model_note = MiscNotesModifications._from_tag(note, inst.id)
                if model_note is not None:
                    inst.notes.append(model_note)
        return inst


class NoteBase(object):
    _tag_name = "misc_notes"

    id = Column(Integer, primary_key=True)
    text = Column(UnicodeText)

    def __str__(self):
        return self.text


class MiscNotesModifications(Base, NoteBase):
    __tablename__ = "MiscNotesModifications"

    modification_id = Column(Integer, ForeignKey(Modification.id), index=True)

    @classmethod
    def _from_tag(cls, tag, modification_id):
        if tag.text is None:
            return
        return cls(text=tag.text, modification_id=modification_id)


class Specificity(Base):
    __tablename__ = "Specificity"

    _tag_name = "specificity_row"

    id = Column(Integer, primary_key=True)
    position_id = Column(Integer, ForeignKey(Position.id), index=True)
    classification_id = Column(Integer, ForeignKey(Classification.id), index=True)
    classification = relationship("Classification", uselist=False)
    # Map through one_letter
    amino_acid = Column(Unicode(10), ForeignKey(AminoAcid.one_letter), index=True)
    modification_id = Column(Integer, ForeignKey(Modification.id), index=True)
    hidden = Column(Boolean, index=True)
    group = Column(Integer, index=True)
    neutral_losses = relationship("SpecificityToNeutralLoss")

    notes = relationship("MiscNotesSpecificities")

    position = relationship(Position)

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            position_id=int(attrib['position_key']),
            classification_id=int(attrib["classifications_key"]),
            group=int(attrib['spec_group']),
            hidden=bool(int(attrib["hidden"])),
            amino_acid=attrib["one_letter"],
            modification_id=int(attrib["mod_key"]),
            )
        for note in tag:
            if note.tag == MiscNotesSpecificities._tag_name:
                model_note = MiscNotesSpecificities._from_tag(note, inst.id)
                if model_note is not None:
                    inst.notes.append(model_note)
        return inst


class MiscNotesSpecificities(Base, NoteBase):
    __tablename__ = "MiscNotesSpecificities"

    specificity_id = Column(Integer, ForeignKey(Specificity.id), index=True)

    @classmethod
    def _from_tag(cls, tag, specificity_id):
        if tag.text is None:
            return
        return cls(text=tag.text, specificity_id=specificity_id)


class NeutralLoss(Base):
    __tablename__ = "NeutralLoss"

    _tag_name = "neutral_losses_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            brick_string=(attrib['brick']),
            count=int(attrib["num_brick"]),
            specificity_id=int(attrib["spec_key"])
            )
        return inst

    id = Column(Integer, primary_key=True)
    brick_string = Column(Unicode(64), index=True)
    specificity_id = Column(Integer, ForeignKey(Specificity.id), index=True)
    count = Column(Integer)


@has_composition("_composition")
class SpecificityToNeutralLoss(Base):
    __tablename__ = "SpecificityToNeutralLoss"

    _tag_name = "spec2nl_row"

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls(
            id=int(attrib['record_id']),
            specificity_id=int(attrib["spec_key"]),
            monoisotopic_mass=float(attrib["nl_mono_mass"]),
            average_mass=float(attrib["nl_avge_mass"]),
            is_required_peptide_neutral_loss=bool(int(attrib["is_req_pep_nl"])),
            is_peptide_neutral_loss=bool(int(attrib["is_pep_nl"])),
            is_slave=bool(int(attrib["is_slave_nl"])),
            _composition=attrib['nl_composition']
            )
        return inst

    id = Column(Integer, primary_key=True)
    specificity_id = Column(Integer, ForeignKey(Specificity.id), index=True)
    specificity = relationship(Specificity, uselist=False)
    monoisotopic_mass = Column(Numeric(12, 6, asdecimal=False), index=True)
    average_mass = Column(Numeric(12, 6, asdecimal=False), index=True)
    _composition = Column(Unicode(128))
    is_slave = Column(Boolean, index=True)
    is_peptide_neutral_loss = Column(Boolean, index=True)
    is_required_peptide_neutral_loss = Column(Boolean, index=True)


class CrossreferenceSource(Base):
    __tablename__ = "CrossreferenceSource"
    _tag_name = "xref_sources_row"

    id = Column(Integer, primary_key=True)
    source = Column(Unicode(64), index=True)

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls()
        inst.id = int(attrib['record_id'])
        inst.source = attrib['xref_source']
        return inst


class Crossreference(Base):
    __tablename__ = "Crossreference"

    _tag_name = "xrefs_row"

    id = Column(Integer, primary_key=True)
    source_id = Column(Integer, ForeignKey(CrossreferenceSource.id), index=True)
    source = relationship(CrossreferenceSource, uselist=False)
    url = Column(Unicode(128))
    modification_id = Column(Integer, ForeignKey(Modification.id), index=True)
    text = Column(UnicodeText)

    @classmethod
    def from_tag(cls, tag):
        attrib = tag.attrib
        inst = cls()
        inst.id = int(attrib['record_id'])
        inst.url = attrib['xref_url']
        inst.source_id = int(attrib['xref_source_key'])
        inst.modification_id = int(attrib["mod_key"])
        text = []
        for node in tag.getchildren():
            if node.tag == "xref_text":
                if node.text is not None:
                    text.append(node.text)
        inst.text = "\n".join(text)
        return inst


def create(doc_path, output_path="sqlite://"):
    '''
    Parse the relational table-like XML file provided by http://www.unimod.org/downloads.html
    and convert each <tag>_row into an equivalent database entry.

    By default the table will be held in memory.
    '''
    tree = preprocess_xml(doc_path)
    engine = create_engine(output_path)
    Base.metadata.create_all(engine)
    session = sessionmaker(bind=engine, autoflush=False)()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=sa_exc.SAWarning)
        for model in Base._decl_class_registry.values():
            if hasattr(model, "_tag_name") and hasattr(model, "from_tag"):
                for tag in tree.iterfind(".//" + model._tag_name):
                    session.add(model.from_tag(tag))
            session.commit()
    return session


def session(path="sqlite:///unimod.db"):
    engine = create_engine(path)
    Base.metadata.create_all(engine)
    session = sessionmaker(bind=engine, autoflush=False)()
    return session


class Unimod(object):
    def __init__(self, path=None):
        if path is None:
            self.path = None
            self.session = create(_unimod_xml_download_url)
        else:
            self.path = path
            try:
                self.session = session(path)
                if self.session.query(Modification).first() is None:
                    raise Exception()
            except:
                # Database may not yet exist at that location
                self.session = create(_unimod_xml_download_url, path)
                self.session.query(Modification).first()

    def get(self, identifier, strict=True):
        if isinstance(identifier, int):
            mod = self.session.query(Modification).get(identifier)
            if mod is None:
                raise KeyError(identifier)
            return mod
        elif isinstance(identifier, basestring):
            if strict:
                mod = self.session.query(Modification).filter(
                    (Modification.full_name == identifier) |
                    (Modification.code_name == identifier) |
                    (Modification.ex_code_name == identifier)).first()
                if mod is None:
                    alt_name = self.session.query(AlternativeName).filter(
                        AlternativeName.alt_name == identifier).first()
                    if alt_name is None:
                        raise KeyError(identifier)
                    mod = alt_name.modification
                return mod
            else:
                qname = "%%%s%%" % identifier
                mod = self.session.query(Modification).filter(
                    (Modification.full_name.like(qname)) |
                    (Modification.code_name.like(qname)) |
                    (Modification.ex_code_name.like(qname))).first()
                if mod is None:
                    alt_name = self.session.query(AlternativeName).filter(
                        AlternativeName.alt_name.like(qname)).first()
                    if alt_name is None:
                        raise KeyError(identifier)
                    mod = alt_name.modification
                return mod

    by_title = by_name = get

    __getitem__ = get

    @property
    def mods(self):
        return self.session.query(Modification).all()

    def __iter__(self):
        return iter(self.session.query(Modification).yield_per(1000))


def load(path=None):
    return Unimod(path)


def unimod_modification_to_dict(mod):
    spec = {
        "approved": mod.approved,
        "avge_mass": mod.average_mass,
        "composition": mod.composition,
        "full_name": mod.full_name,
        "mono_mass": mod.monoisotopic_mass,
        "record_id": mod.id,
        "specificity": [
            unimod_specificity_to_dict(s) for s in mod.specificities
        ],
        "title": mod.code_name,
        "alt_names": mod.alternative_names + [mod.ex_code_name]
    }
    return spec


def unimod_specificity_to_dict(specificity):
    spec = {
        "classification": str(specificity.classification),
        "hidden": specificity.hidden,
        "position": str(specificity.position),
        "site": specificity.amino_acid,
        "spec_group": specificity.group
    }
    spec['neutral_losses'] = [
        unimod_neutral_loss_to_dict(nl) for nl in specificity.neutral_losses
    ]
    return spec


def unimod_neutral_loss_to_dict(neutral_loss):
    spec = {
        "composition": neutral_loss.composition,
        "mono_mass": neutral_loss.monoisotopic_mass,
    }
    return spec


if __name__ == "__main__":
    import sys
    import json

    db = Unimod()
    mods = db.mods
    encoded = [unimod_modification_to_dict(mod) for mod in mods]
    encoded.sort(key=lambda x: x["record_id"])
    # need method to impose patches
    try:
        fname = sys.argv[1]
        with open(fname, 'wt') as fh:
            json.dump(encoded, fh, sort_keys=True, indent=4)
    except IndexError:
        json.dump(encoded, sys.stdout, sort_keys=True, indent=4)



import unittest
import warnings
from io import BytesIO

from glycopeptidepy.io import fasta

from .common import datafile


class TestFastaIO(unittest.TestCase):
    def open_stream(self):
        return open(datafile("proteins.fa"), 'rb')

    def test_fasta_parser(self):
        n_proteins_in_fasta = 243
        with self.open_stream() as f:
            parser = fasta.ProteinFastaFileReader(f, index=True)
            assert len(parser.index) == n_proteins_in_fasta
            i = 0
            for protein in parser:
                i += 1
                self.assertEqual(protein.name.db, 'sp')
            assert i == n_proteins_in_fasta
            protein = parser['sp|Q9NYU2|UGGG1_HUMAN UDP-glucose:glycoprotein glucosyltransferase 1']
            assert str(protein).endswith("KREEL")

    def test_fasta_writer(self):
        outbuffer = BytesIO()
        writer = fasta.FastaFileWriter(outbuffer)
        writer.write("description", "PEPTIDE")
        writer.write("description2", "PROTEIN")
        content = outbuffer.getvalue()
        reference = b'''>description
PEPTIDE

>description2
PROTEIN

'''
        assert content == reference
        with self.open_stream() as f:
            parser = fasta.ProteinFastaFileReader(f)
            proteins = list(parser)
        outbuffer = BytesIO()
        writer = fasta.ProteinFastaFileWriter(outbuffer)
        writer.writelines(proteins)
        outbuffer.seek(0)
        new_proteins = list(fasta.ProteinFastaFileReader(outbuffer))
        assert proteins == new_proteins

    def test_replaces_unknown(self):
        reader = fasta.ProteinFastaFileReader(
            datafile("unknown_amino_acid.fa"), replace_unknown=True)
        contents = list(reader)
        assert len(contents) == 1
        reader = fasta.ProteinFastaFileReader(
            datafile("unknown_amino_acid.fa"), replace_unknown=False)
        with warnings.catch_warnings(record=True) as warning_log:
            warnings.simplefilter("always")
            should_fail = list(reader)
            assert len(should_fail) == 0
            assert len(warning_log) == 1
            assert "B" in str(warning_log[0].message)


class TestHeaderParsing(unittest.TestCase):
    def test_proper_uniprot(self):
        header = ">sp|P16066|ANPRA_HUMAN Atrial natriuretic peptide receptor 1"
        result = fasta.default_parser(header)
        self.assertEqual(result.accession, "P16066")

    def test_without_prefix(self):
        header = "sp|P16066|ANPRA_HUMAN Atrial natriuretic peptide receptor 1"
        result = fasta.default_parser(header)
        self.assertEqual(result.accession, "P16066")

    def test_partial_uniprot(self):
        header = "P16066|ANPRA_HUMAN"
        result = fasta.default_parser(header)
        self.assertEqual(result.accession, "P16066")

    def test_accession_alone(self):
        header = "P16066"
        result = fasta.default_parser(header)
        self.assertRaises(AttributeError, lambda: result.accession)
        self.assertEqual(result[0], "P16066")


class TestPEFF(unittest.TestCase):
    def open_stream(self, mode='rb'):
        return open(datafile("SmallTestDB-PEFF1.0.peff"), mode)

    def test_peff_parser(self):
        with self.open_stream() as f:
            parser = fasta.PEFFReader(f)
            assert parser.number_of_entries == 29
            proteins = list(parser)
            n = len(proteins)

            assert n == parser.number_of_entries
            annotations = proteins[0].annotations
            assert len(annotations) == 7
            assert annotations['Length'] == 529


class DictWrapperSuiteBase(object):
    def _make_instance(self):
        raise NotImplementedError()

    def _make_reference(self):
        raise NotImplementedError()

    def test_getitem(self):
        inst = self._make_instance()
        ref = self._make_reference()
        for key in ref:
            assert inst[key] == ref[key]

    def test_getattr(self):
        inst = self._make_instance()
        ref = self._make_reference()
        for key in ref:
            assert getattr(inst, key) == ref[key]

    def test_contains(self):
        inst = self._make_instance()
        ref = self._make_reference()
        for key in ref:
            assert key in inst

    def test_keys(self):
        inst = self._make_instance()
        ref = self._make_reference()
        assert set(inst.keys()) == set(ref.keys())

    def test_values(self):
        inst = self._make_instance()
        ref = self._make_reference()
        assert sorted(inst.values(), key=str) == sorted(ref.values(), key=str)

    def test_items(self):
        inst = self._make_instance()
        ref = self._make_reference()
        assert sorted(inst.items()) == sorted(ref.items())

    def test_len(self):
        inst = self._make_instance()
        ref = self._make_reference()
        assert len(inst) == len(ref)


class TestFastaHeader(DictWrapperSuiteBase, unittest.TestCase):
    defline = ">sp|Q9NYU2|UGGG1_HUMAN UDP-glucose:glycoprotein glucosyltransferase 1"

    def _make_instance(self):
        return fasta.default_parser(self.defline)

    def _make_reference(self):
        ref = {'accession': 'Q9NYU2',
               'db': 'sp',
               'description': 'UDP-glucose:glycoprotein glucosyltransferase 1',
               'name': 'UGGG1_HUMAN'}
        return ref


class TestPEFFHeaderBlock(DictWrapperSuiteBase, unittest.TestCase):
    block = b'''# PEFF 1.0
# GeneralComment=this is a hand crafted peff file
# GeneralComment=This is a second comment line.
# //
# DbName=UniProtKB/Swiss-Prot-extract
# Prefix=sp
# DbDescription=selected entries from UniProtKB
# Decoy=false
# DbSource=ftp://ftp.uniprot.org//ftp/pub/databases/uniprot/current_release/knowledgebase/complete/uniprot_sprot.dat.gz
# DbVersion=55.5
# DbDate=20080610
# NumberOfEntries=28
# Conversion=manual extraction of entries, conversion with prepareDb.pl and then manually edited
# SequenceType=AA
# SpecificKey=3D-Status:status of a 3-D structure
# SpecificValue=3D-Status:(available|unsure|not available)
# SpecificKey=isoform:description of a specific isoform
# SpecificValue=isoform:(xsd:type=string)
# //'''

    def _make_instance(self):
        buff = BytesIO(self.block)
        reader = fasta.PEFFReader(buff)
        return reader.blocks[0]

    def _make_reference(self):
        return {u'Conversion': u'manual extraction of entries, conversion with prepareDb.pl and then manually edited',
                u'DbDate': u'20080610',
                u'DbDescription': u'selected entries from UniProtKB',
                u'DbName': u'UniProtKB/Swiss-Prot-extract',
                u'DbSource': (u'ftp://ftp.uniprot.org//ftp/pub/databases/uniprot/'
                              u'current_release/knowledgebase/complete/uniprot_sprot.dat.gz'),
                u'DbVersion': u'55.5',
                u'Decoy': u'false',
                u'NumberOfEntries': u'28',
                u'Prefix': u'sp',
                u'SequenceType': u'AA',
                u'SpecificKey': [u'3D-Status:status of a 3-D structure',
                                 u'isoform:description of a specific isoform'],
                u'SpecificValue': [u'3D-Status:(available|unsure|not available)',
                                   u'isoform:(xsd:type=string)']}


if __name__ == '__main__':
    unittest.main()



import unittest

from glycopeptidepy.utils.intervals import Interval


iv0 = Interval(0, 10)
iv1 = Interval(-10, -5)
iv2 = Interval(-10, 0)
iv3 = Interval(-10, 5)
iv4 = Interval(-10, 10)
iv5 = Interval(-10, 20)
iv6 = Interval(0, 20)
iv7 = Interval(5, 20)
iv8 = Interval(10, 20)
iv9 = Interval(15, 20)
iv10 = Interval(-5, 0)


class TestIntervals(unittest.TestCase):
    def test_interval_overlaps_interval(self):
        assert iv0.overlaps(iv0)
        assert not iv0.overlaps(iv1)
        assert iv0.overlaps(iv2)
        assert iv0.overlaps(iv3)
        assert iv0.overlaps(iv4)
        assert iv0.overlaps(iv5)
        assert iv0.overlaps(iv6)
        assert iv0.overlaps(iv7)
        assert iv0.overlaps(iv8)
        assert not iv0.overlaps(iv9)

    def test_contains_interval(self):
        assert iv0.contains_interval(iv0)
        assert not iv0.contains_interval(iv1)
        assert not iv0.contains_interval(iv2)
        assert not iv0.contains_interval(iv3)
        assert not iv0.contains_interval(iv4)
        assert not iv0.contains_interval(iv5)
        assert not iv0.contains_interval(iv6)
        assert not iv0.contains_interval(iv7)
        assert not iv0.contains_interval(iv8)
        assert not iv0.contains_interval(iv9)
        assert not iv0.contains_interval(iv10)

        assert not iv2.contains_interval(iv0)
        assert iv2.contains_interval(iv1)
        assert iv2.contains_interval(iv2)
        assert not iv2.contains_interval(iv3)
        assert not iv2.contains_interval(iv4)
        assert not iv2.contains_interval(iv5)
        assert not iv2.contains_interval(iv6)
        assert not iv2.contains_interval(iv7)
        assert not iv2.contains_interval(iv8)
        assert not iv2.contains_interval(iv9)
        assert iv2.contains_interval(iv10)


if __name__ == '__main__':
    unittest.main()



from .glycosylated_sequence import (
    find_glycosaminoglycan_sequons,
    find_n_glycosylation_sequons,
    find_o_glycosylation_sequons,
    _n_glycosylation,
    _o_glycosylation,
    _gag_linker_glycosylation)

from .implementation import (
    PeptideSequence,
    NamedSequence, AnnotatedSequence,
    ProteinSequence,
    list_to_sequence,)

from .base import PeptideSequenceBase

parse = PeptideSequence


__all__ = [
    "PeptideSequenceBase",
    "PeptideSequence", "NamedSequence", "ProteinSequence",
    "AnnotatedSequence", "find_glycosaminoglycan_sequons",
    "find_o_glycosylation_sequons", "find_n_glycosylation_sequons",
    "list_to_sequence", "parse",
    "_n_glycosylation", "_o_glycosylation", "_gag_linker_glycosylation",
]



import unittest

from glycopeptidepy.utils import collectiontools


class TestCollectionTools(unittest.TestCase):
    def test_groupby(self):
        in_data = ["SPAM", "SPAM", "MAPS"]
        groups = collectiontools.groupby(in_data)
        assert len(groups) == 2
        assert sum({k: len(v) for k, v in groups.items()}.values()) == len(in_data)
        # test key_fn changing aggregation
        groups = collectiontools.groupby(in_data, key_fn=frozenset)
        assert len(groups) == 1
        assert sum({k: len(v) for k, v in groups.items()}.values()) == len(in_data)

    def test_accumulator_bag(self):
        bag = collectiontools._AccumulatorBag({"a": 4, "b": 3, (1, 2, 3): 2})
        abag = collectiontools._AccumulatorBag({"a": 8, "b": 6, (1, 2, 3): 4})
        bbag = bag + bag
        assert abag == bbag
        assert len(bag) == 3
        assert "a" in bag
        assert (1, 2, 3) in bag
        assert list(bag) == list(bag.keys())
        assert sum(bag.values()) == 4 + 3 + 2


if __name__ == '__main__':
    unittest.main()



import unittest

from glycopeptidepy.structure.sequence.implementation import PeptideSequence

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from glycopeptidepy import enzyme
from glycopeptidepy.io.fasta import ProteinFastaFileParser

_protein_fasta = '''
>sp|Q9Y251|HPSE_HUMAN Heparanase OS=Homo sapiens GN=HPSE PE=1 SV=2
MLLRSKPALPPPLMLLLLGPLGPLSPGALPRPAQAQDVVDLDFFT
QEPLHLVSPSFLSVTIDANLATDPRFLILLGSPKLRTLARGLSPA
YLRFGGTKTDFLIFDPKKESTFEERSYWQSQVNQDICKYGSIPPD
VEEKLRLEWPYQEQLLLREHYQKKFKNSTYSRSSVDVLYTFANCS
GLDLIFGLNALLRTADLQWNSSNAQLLLDYCSSKGYNISWELGNE
PNSFLKKADIFINGSQLGEDFIQLHKLLRKSTFKNAKLYGPDVGQ
PRRKTAKMLKSFLKAGGEVIDSVTWHHYYLNGRTATKEDFLNPDV
LDIFISSVQKVFQVVESTRPGKKVWLGETSSAYGGGAPLLSDTFA
AGFMWLDKLGLSARMGIEVVMRQVFFGAGNYHLVDENFDPLPDYW
LSLLFKKLVGTKVLMASVQGSKRRKLRVYLHCTNTDNPRYKEGDL
TLYAINLHNVTKYLRLPYPFSNKQVDKYLLRPLGPHGLLSKSVQL
NGLTLKMVDDQTLPPLMEKPLRPGSSLGLPAFSYSFFVIRNAKVA
ACI

>sp|P02763|A1AG1_HUMAN Alpha-1-acid glycoprotein 1 OS=Homo sapiens GN=ORM1 PE=1 SV=1
MALSWVLTVLSLLPLLEAQIPLCANLVPVPITNATLDQITGKWFYIASAFRNEEYNKSVQ
EIQATFFYFTPNKTEDTIFLREYQTRQDQCIYNTTYLNVQRENGTISRYVGGQEHFAHLL
ILRDTKTYMLAFDVNDEKNWGLSVYADKPETTKEQLGEFYEALDCLRIPKSDVVYTDWKK
DKCEPLEKQHEKERKQEEGES

'''

parser = ProteinFastaFileParser(StringIO(_protein_fasta))

heparanase = next(parser)
agp1 = next(parser)


class TestProtease(unittest.TestCase):
    def test_digest(self):
        trypsin = enzyme.Protease("trypsin")
        example_found = False
        result = trypsin.cleave(heparanase, 2)
        seq = str(heparanase)
        assert len(result) == 177
        for peptide, start, stop, missed in result:
            assert missed < 3
            assert missed == trypsin.missed_cleavages(peptide)
            assert seq[start:stop] == peptide
            if peptide == "KFKNSTYSR":
                example_found = True
        if not example_found:
            raise AssertionError("Did not produce overlapped digest")

    def test_semispecific(self):
        trypsin = enzyme.Protease("trypsin")
        c_term_example = False
        n_term_example = False
        seq = str(agp1)
        result = trypsin.cleave(agp1, semispecific=True, missed_cleavages=2)
        assert len(result) == 1037
        for peptide, start, end, missed in result:
            assert missed < 3
            assert missed == trypsin.missed_cleavages(peptide)
            assert seq[start:end] == peptide
            if peptide == "LVPVPITNATLDQITGK":
                c_term_example = True
            if peptide == 'ENGT':
                n_term_example = True

        if not (n_term_example and c_term_example):
            raise AssertionError("Did not produce overlapped digest")

    def test_combine(self):
        trypsin = enzyme.Protease("trypsin")
        gluc = enzyme.Protease("glutamyl endopeptidase")
        seq = "PEPTIDER"
        self.assertEqual(trypsin.missed_cleavages(seq), 0)
        self.assertEqual(gluc.missed_cleavages(seq), 2)
        merged = enzyme.Protease.combine(trypsin, gluc)
        self.assertEqual(merged.missed_cleavages(seq), 2)


if __name__ == '__main__':
    unittest.main()



'''A REST API for the UniProt protein database and a parser for their XML format.
'''
import csv
import json
import warnings
import threading
import gzip
import io
import logging

from typing import IO, List, Optional, ClassVar, Dict
from urllib.request import urlopen
from collections import defaultdict

import requests

from lxml import etree

from glypy.utils import make_struct
from glypy.io.glyspace import UniprotRDFClient

from glycopeptidepy.utils import simple_repr
from glycopeptidepy.structure.glycan import GlycosylationType

from six import add_metaclass, string_types as basestring

URI_TEMPLATE = "https://www.uniprot.org/uniprot/{accession}.xml"
BATCH_URI = "https://www.uniprot.org/uploadlists/"
NSMAP = {"up": "http://uniprot.org/uniprot"}
# BATCH_URI_QUERY = "https://rest.uniprot.org/uniprotkb/search?format=xml&query="
BATCH_URI_QUERY = "https://rest.uniprot.org/uniprotkb/accessions?format=xml&accessions="

VERIFY_SSL = False

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _batch_uri_query_builder(accessions: List[str]) -> str:
    # query = " OR ".join([f"accession:{acc}" for acc in accessions])
    query = ','.join(accessions)
    return BATCH_URI_QUERY + query


class UniProtFeatureMeta(type):
    _cache: ClassVar[Dict[str, 'UniProtFeatureMeta']] = {}

    def __new__(cls, name, parents, attrs):
        new_type = type.__new__(cls, name, parents, attrs)
        try:
            cls._cache[new_type.feature_type] = new_type
        except AttributeError:
            pass
        return new_type

    def from_feature_type(self, feature_type: str) -> 'UniProtFeatureMeta':
        return self._cache[feature_type]

    def handle_tag(self, tag: etree.Element) -> 'UniProtFeatureBase':
        feature_type = tag.attrib['type']
        try:
            feature_type_t = self.from_feature_type(feature_type)  # pylint: disable=no-value-for-parameter
            return feature_type_t.fromxml(tag)
        except KeyError:
            return None


@add_metaclass(UniProtFeatureMeta)
class UniProtFeatureBase(object):
    __repr__ = simple_repr

    known: bool = True
    _name: Optional[str] = None

    @property
    def is_defined(self):
        return self.known

    def describe(self) -> str:
        return self.description.split("; ")

    @property
    def name(self):
        return self.feature_type


class Keyword(str):
    id: str

    def __new__(cls, content, idstr):
        obj = str.__new__(cls, content)
        obj.id = idstr
        return obj

    def __reduce__(self):
        return self.__class__, (str(self), self.id)


class PeptideBase(UniProtFeatureBase):
    start: int
    end: int

    def __init__(self, start, end, known=True):
        self.start = start
        self.end = end
        self.known = known

    @classmethod
    def fromxml(cls, feature: etree.Element):
        known = True

        try:
            begin = int(feature.find(".//{http://uniprot.org/uniprot}begin", NSMAP).attrib['position']) - 1
        except (KeyError, AttributeError):
            begin = 0
            known = False
        try:
            end = int(feature.find(".//{http://uniprot.org/uniprot}end", NSMAP).attrib['position'])
        except (KeyError, AttributeError):
            end = float('inf')
            known = False
        if not known:
            if begin == 0 and end == float('inf'):
                begin = int(feature.find(
                    ".//{http://uniprot.org/uniprot}location", NSMAP).attrib['position']) - 1
                end = begin + 1
                known = True
        return cls(begin, end, known)


class SignalPeptide(PeptideBase):
    feature_type = 'signal peptide'


class Propeptide(PeptideBase):
    feature_type = 'propeptide'


class TransitPeptide(PeptideBase):
    feature_type = "transit peptide"


class Peptide(PeptideBase):
    feature_type = 'peptide'


class MatureProtein(PeptideBase):
    feature_type = 'mature protein'


class SpanBase(UniProtFeatureBase):
    feature_type = "__spanbase__"

    start: int
    end: int

    def __init__(self, name, start, end, known=True):
        self.name = name
        self.start = start
        self.end = end
        self.known = known

    @classmethod
    def fromxml(cls, feature: etree.Element):
        known = True
        try:
            try:
                begin = int(feature.find(
                    ".//up:begin", NSMAP).attrib['position']) - 1
            except KeyError:
                begin = 0
                known = False
            try:
                end = int(feature.find(
                    ".//up:end", NSMAP).attrib['position'])
            except KeyError:
                end = float('inf')
                known = False
        except AttributeError:
            begin = int(feature.find(
                ".//up:position", NSMAP).attrib['position']) - 1
            end = begin + 1

        description = feature.attrib.get("description")
        return cls(description, begin, end, known)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def description(self):
        return self.name

    def __repr__(self):
        repr_str = super(SpanBase, self).__repr__()
        prefix, suffix = repr_str.split("(", 1)
        augmented = "%s(name=%r, %s" % (prefix, self.name, suffix)
        return augmented


class Domain(SpanBase):
    feature_type = 'domain'


class Chain(SpanBase):
    feature_type = 'chain'


class RegionOfInterest(Domain):
    feature_type = 'region of interest'


class InitiatorMethionine(Domain):
    feature_type = "initiator methionine"


class DisulfideBond(Domain):
    feature_type = "disulfide bond"

    @classmethod
    def fromxml(cls, feature):
        name = "disulfide bond"
        known = True
        try:
            try:
                begin = int(feature.find(
                    ".//up:begin", NSMAP).attrib['position']) - 1
            except KeyError:
                begin = 0
                known = False
            try:
                end = int(feature.find(
                    ".//up:end", NSMAP).attrib['position'])
            except KeyError:
                end = float('inf')
                known = False
        except AttributeError:
            begin = end = int(feature.find(
                ".//up:position", NSMAP).attrib['position'])
        return cls(name, begin, end, known)


class ModifiedResidue(UniProtFeatureBase):
    feature_type = 'modified residue'

    position: int
    modification: str

    def __init__(self, position, modification):
        self.position = position
        self.modification = modification

    @property
    def name(self):
        return self.modification

    @property
    def description(self):
        return self.modification

    @property
    def start(self):
        return self.position

    @property
    def end(self):
        return self.position + 1

    @classmethod
    def fromxml(cls, feature):
        return cls(
            int(feature.find(
                ".//up:position", NSMAP).attrib['position']) - 1,
            feature.attrib["description"])


class SequenceVariant(UniProtFeatureBase):
    feature_type = "sequence variant"

    position: int
    original: str
    variation: str

    def __init__(self, position, original, variation):
        self.position = position
        self.original = original
        self.variation = variation

    @property
    def start(self):
        return self.position

    @property
    def end(self):
        return self.position + 1

    @classmethod
    def fromxml(cls, feature):
        position = feature.find(".//up:position", NSMAP)
        if position is None:
            # Deletion mutations not supported
            return None
        else:
            position = int(position.attrib['position']) - 1
            original = feature.find(".//up:original", NSMAP)
            if original is not None:
                original = original.text
            variation = feature.find(".//up:variation", NSMAP)
            if variation is not None:
                variation = variation.text
        return cls(position, original, variation)

    @property
    def description(self):
        return "%s->%s" % (self.original, self.variation)


class Site(Domain):
    feature_type = 'site'


class GlycosylationSite(UniProtFeatureBase):
    feature_type = 'glycosylation site'

    position: int
    glycosylation_type: str
    description: str

    def __init__(self, position, glycosylation_type, description=None):
        self.position = position
        try:
            self.glycosylation_type = GlycosylationType[glycosylation_type]
        except KeyError:
            self.glycosylation_type = GlycosylationType[None]
        self.description = description

    @property
    def name(self):
        return self.glycosylation_type

    @property
    def start(self):
        return self.position

    @property
    def end(self):
        return self.position + 1

    @classmethod
    def fromxml(cls, feature):
        position = int(feature.find(
            ".//up:position", NSMAP).attrib['position']) - 1
        glycosylation_type = feature.attrib["description"].split(" ")[0]
        return cls(position, glycosylation_type, feature.attrib["description"])



# @dataclass
# class UniProtProtein:
#     sequence: str
#     features: List[UniProtFeatureBase]
#     recommended_name: str
#     gene_name: str
#     names: List[str]
#     accessions: List[str]
#     keywords: List[Keyword]
#     dbreferences: List


_UniProtProtein = make_struct("UniProtProtein", (
    "sequence", "features", "recommended_name", "gene_name", "names", "accessions",
    "keywords", "dbreferences"))


class UniProtProtein(_UniProtProtein):
    __slots__ = ()

    def features_of_type(self, feature_type):
        if isinstance(feature_type, basestring):
            return [feature for feature in self.features if feature.feature_type == feature_type]
        elif isinstance(feature_type, type):
            return [feature for feature in self.features if isinstance(feature, feature_type)]
        else:
            raise TypeError("Can't use object of type %r, %r" % (type(feature_type), feature_type))

    @property
    def chains(self):
        return self.features_of_type(Chain)

    @property
    def domains(self):
        return self.features_of_type(Domain)

    @property
    def variants(self):
        return self.features_of_type(SequenceVariant)

    @property
    def peptides(self):
        return self.features_of_type(PeptideBase)


def get_etree_for(accession):
    tree = etree.parse(_open_url_for(accession))
    return tree


def _open_url_for(accession):
    return urlopen(URI_TEMPLATE.format(accession=accession))


def parse(tree: etree._Element, error=False) -> UniProtProtein:
    '''Parse an XML document into a :class:`UniProtProtein`.

    Parameters
    ----------
    tree: lxml.etree.Element
        The root XML element of the tree to parse.
    errors: bool
        Whether to treat errors during feature parsing
        as warnings or exceptions.

    Returns
    -------
    UniProtProtein
    '''
    entry = tree.find(".//up:entry", NSMAP)
    if entry is None:
        if hasattr(tree, "tag") and tree.tag.endswith("entry"):
            entry = tree
        else:
            raise ValueError("Could not find root entry!")
    seq = entry.find(
        "./up:sequence", NSMAP).text
    if seq is not None:
        seq = seq.replace("\n", '')
    names = [el.text for el in entry.findall(
        ".//up:protein/*/up:fullName", NSMAP)] + [el.text for el in entry.findall(
            ".//up:protein/*/up:shortName", NSMAP)]
    recommended_name_tags = entry.findall(
        ".//up:protein/up:recommendedName", NSMAP) + entry.findall(".//up:protein/*/up:recommendedName", NSMAP)
    if recommended_name_tags:
        recommended_name_tag = recommended_name_tags[0]
    else:
        recommended_name_tag = None
    if recommended_name_tag is not None:
        if recommended_name_tag.text.strip():
            recommended_name = recommended_name_tag.text.strip()
        else:
            recommended_name = (recommended_name_tag[0].text)
    else:
        try:
            recommended_name = names[0]
        except IndexError:
            recommended_name = ""
    gene_name_tag = entry.find(".//up:name", NSMAP)
    if gene_name_tag is not None:
        gene_name = gene_name_tag.text
    else:
        gene_name = ""
    accessions = [el.text for el in entry.findall(
        ".//up:accession", NSMAP)]
    dbreferences = defaultdict(list)
    for tag in entry.findall(".//up:dbReference", NSMAP):
        db = tag.attrib['type']
        rec = dict(id=tag.attrib['id'])
        for prop in tag.findall(".//up:property", NSMAP):
            rec[prop.attrib['type']] = prop.attrib['value']
        dbreferences[db].append(rec)

    features = []
    exc_type = Exception
    for tag in entry.findall(".//up:feature", NSMAP):
        feature_type = tag.attrib['type']
        try:
            feature_obj = UniProtFeatureBase.handle_tag(tag)
            if feature_obj is not None:
                features.append(feature_obj)
        except exc_type as e:
            if error:
                raise
            else:
                warnings.warn("An exception %r occurred while parsing feature type %s for %s" % (
                    e, feature_type, accessions[0]))
    keywords = set()
    for kw in entry.findall(".//up:keyword", NSMAP):
        keywords.add(Keyword(kw.text, kw.attrib['id']))
    return UniProtProtein(
        seq, features, recommended_name, gene_name, names, accessions,
        keywords, dbreferences)


def parse_all(tree, error=False):
    """Parse an XML document containing multiple <up:entry> elements into
    a list of :class:`UniProtProtein` instances.

    Parameters
    ----------
    tree : lxml.etree.Element
        The root element of the XML document to parse.
    errors: bool
        Whether to treat errors during feature parsing
        as warnings or exceptions.

    Returns
    -------
    list
    """
    entries = tree.findall(".//up:entry", NSMAP)
    return [parse(e, error=error) for e in entries]


def iterparse(stream, error=False):
    for event, tag in etree.iterparse(stream, huge_tree=True, events=("start", "end")):
        if event == "end":
            if tag.tag.split("}")[1] == 'entry':
                rec = parse(tag, error=error)
                tag.clear()
                yield rec


def is_uniprot_xml(stream: IO) -> bool:
    start = None
    if isinstance(stream, io.BufferedReader) or hasattr('peek'):
        peek = stream.peek(100)
    elif stream.seekable():
        start = stream.tell()
        peek = stream.read(100)
    else:
        peek = stream.read(100)
    if isinstance(peek, str):
        peek = peek.encode('utf8')
    result = b"<uniprot " in peek
    if start is not None:
        stream.seek(start)
    return result


def get_features_for(accession, error=False):
    tree = get_etree_for(accession)
    return parse(tree, error)


def get_features_for_many(accessions, error=False):
    query_to_doc = dict()
    url = _batch_uri_query_builder(accessions)
    r = requests.get(url, verify=VERIFY_SSL)
    tree = etree.fromstring(r.content)
    for doc in parse_all(tree, error=error):
        for acc in doc.accessions:
            query_to_doc[acc] = doc
    results = []
    for acc in accessions:
        if acc in query_to_doc:
            results.append((acc, query_to_doc[acc]))
        else:
            results.append((acc, None))
    return results


def get(accessions):
    if accessions is None:
        raise TypeError("accessions cannot be `None`")
    if isinstance(accessions, basestring):
        return get_features_for(accessions)
    elif len(accessions) < 5:
        return [b for a, b in get_features_for_many(accessions) if b is not None]
    else:
        chunk_size = min(len(accessions) // 5, 15)
        return ProteinDownloader.download(accessions, chunk_size)


class ProteinDownloader(object):  # pragma: no cover
    @staticmethod
    def chunk(seq, n=5):
        i = 0
        while i < len(seq):
            yield seq[i:i + n]
            i += n

    @staticmethod
    def fetch(chunk):
        jobs = []
        results = []

        def task(name):
            try:
                feat = get_features_for(name)
                results.append(feat)
            except Exception as e:
                print(e, name)

        for c in chunk:
            th = threading.Thread(target=task, args=(c,))
            th.start()
            jobs.append(th)

        for th in jobs:
            th.join()

        return results

    @classmethod
    def download(cls, accession_iterable, chunk_size=15, ordered=True):
        accession_iterable = list(accession_iterable)
        chunks = cls.chunk(accession_iterable, chunk_size)
        proteins = []
        for chunk in chunks:
            proteins.extend(cls.fetch(chunk))
        if not ordered:
            return proteins
        acc_map = {}
        for prot in proteins:
            for acc in prot.accessions:
                acc_map[acc] = prot
        proteins = []
        for acc in accession_iterable:
            proteins.append(acc_map.get(acc))
        return proteins

    def __call__(self, *args, **kwargs):
        return self.download(*args, **kwargs)

    @classmethod
    def to_fasta_file(cls, accession_iterable, file_handle, chunk_size=15):
        from .fasta import FastaFileWriter
        proteins = cls.download(accession_iterable, chunk_size=chunk_size)
        writer = FastaFileWriter(file_handle)
        for protein in proteins:
            defline = ">sp|%s|%s %s" % (protein.accessions[0], protein.gene_name, protein.recommended_name)
            defline = defline.replace("\n", " ")
            seq = protein.sequence
            writer.write(defline, seq)
        return file_handle


rdf = UniprotRDFClient()


def search(query):
    url = (
        r"https://rest.uniprot.org/uniprot/search?"
        f"compressed=false&query={query}&format=tsv"
        r"&fields=id,accession,protein_name,organism_name,reviewed,gene_names,annotation_score,length")
    response = requests.get(url, stream=True, verify=VERIFY_SSL)
    response.raise_for_status()
    response_buffer = response.raw
    data = response_buffer.read()
    if response.headers.get("content-encoding") == 'gzip':
        response_buffer = gzip.GzipFile(mode='rb', fileobj=io.BytesIO(data))
        data = response_buffer.read().decode('utf-8')
    else:
        data = data.decode("utf-8")
    reader = csv.reader(data.splitlines(), delimiter='\t')
    header = next(reader)
    return [dict(zip(header, line)) for line in reader]


class _UniProtRestClientBase(object):

    def _buffer_to_tsv(self, data: bytes) -> csv.DictReader:
        return csv.DictReader(io.StringIO(data.decode("utf8")), delimiter='\t')

    def _buffer_to_json(self, data: bytes) -> dict:
        return json.loads(data.decode("utf8"))

    def _get_stream(self, url: str) -> io.IOBase:
        response = requests.get(url, stream=True, verify=VERIFY_SSL)
        response.raise_for_status()
        response_buffer = response.raw
        if response.headers.get("content-encoding") == 'gzip':
            response_buffer = gzip.GzipFile(mode='rb', fileobj=response_buffer)
        return response_buffer

    def _get_response_buffer(self, url: str) -> bytes:
        response = requests.get(url, stream=True, verify=VERIFY_SSL)
        response.raise_for_status()
        response_buffer = response.raw
        data = response_buffer.read()
        if response.headers.get("content-encoding") == 'gzip':
            response_buffer = gzip.GzipFile(
                mode='rb', fileobj=io.BytesIO(data))
            data = response_buffer.read()
        return data



class ProteomeSearchResult(dict):
    bind = None

    def download(self, raw=False, format='fasta'):
        return self.bind.get(self['Proteome ID'], raw=raw, format=format)


class ProteomeClient(_UniProtRestClientBase):

    SearchResultType = ProteomeSearchResult

    def search(self, format='tsv', raw=False, query=None, **kwargs):
        if query is None:
            query = ""
        query += " AND ".join(["%s:%s" % (k, v) for k, v in kwargs.items()])
        uri = "https://rest.uniprot.org/proteomes/search?query={query}&format={format}"
        url = uri.format(query=query, format=format)
        data = self._get_response_buffer(url)
        if raw:
            return io.BytesIO(data)
        if format == 'tsv':
            result = list(map(self.SearchResultType, self._buffer_to_tsv(data)))
            for item in result:
                item.bind = self
            return result
        elif format == 'json':
            return json.loads(data.decode("utf8"))
        elif format == 'list':
            result = []
            for line in data.splitlines():
                r = self.SearchResultType({
                    "Proteome ID": line.strip()
                })
                r.bind = self
                result.append(r)
            return result
        return io.BytesIO(data)

    def get(self, proteome_id, include_isoforms=False, raw=False, format='fasta', stream: bool=False):
        uri = "https://rest.uniprot.org/uniprotkb/stream?{query}&{params}&format={format}"
        query = {"proteome": proteome_id, }
        query = '&'.join(["query=%s:\"%s\"" % (k, v) for k, v in query.items()])
        params = {"includeIsoform": str(bool(include_isoforms)).lower()}
        params = '&'.join(["%s=%s" % (k, v) for k, v in params.items()])
        url = uri.format(query=query, params=params, format=format)
        if stream:
            stream = self._get_stream(url)
            if format == "fasta":
                from .fasta import ProteinFastaFileReader
                return ProteinFastaFileReader(stream, index=False)
            return stream
        data = self._get_response_buffer(url)
        if raw:
            return io.BytesIO(data)
        if format == 'fasta':
            from .fasta import ProteinFastaFileReader
            return ProteinFastaFileReader(io.BytesIO(data))
        elif format == 'xml':
            return etree.fromstring(data)
        else:
            return io.BytesIO(data)

proteome = ProteomeClient()


class TaxonomyClient(_UniProtRestClientBase):

    def search(self, format='tsv', raw=False, query=None, params=None, **kwargs):
        if query is None:
            query = ''
        if params is None:
            params = dict()
        query += "AND".join(["%s:%s" % (k, v)
                                      for k, v in kwargs.items()])
        params = '&'.join(["%s=%s" % (k, v) for k, v in params.items()])
        if params:
            params += '&'
        uri = "https://rest.uniprot.org/taxonomy/search?query={query}&{params}format={format}"
        url = uri.format(query=query, params=params, format=format)
        data = self._get_response_buffer(url)
        if raw:
            return io.BytesIO(data)
        if format == 'tsv':
            result = list(self._buffer_to_tsv(data))
            return result
        if format == 'json':
            return json.loads(data.decode('utf8'))
        elif format == 'list':
            result = data.splitlines()
            return result
        return io.BytesIO(data)


taxonomy = TaxonomyClient()



cimport cython
from glycopeptidepy.structure.modification.descriptors import ModificationCategory
from glypy.utils.cenum cimport EnumValue

cdef class ModificationRuleBase(ModificationBase):
    cpdef bint is_a(self, object category):
        '''Returns whether or not this :class:`ModificationRule` object belongs to
        the specified :class:`~.ModificationCategory`.

        Returns
        -------
        bool
        '''
        return category in self.categories

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(self, ModificationRuleBase):
            temp = self
            self = <ModificationRuleBase>other
            other = temp
        idents = (<ModificationRuleBase>self).names
        if isinstance(other, ModificationRuleBase):
            other_idents = (<ModificationRuleBase>other).names
        else:
            other_idents = {other}
        return bool(idents & other_idents)

    def __ne__(self, other):
        return not self == other

    cpdef basestring serialize(self):
        '''A string representation for inclusion in sequences'''
        return self.name

    @property
    def is_standard(self):
        """Indicates whether the :class:`ModificationRule` describes a modification which
        has been specified from a reference source, and may be fully reconstructed from
        its name alone.

        This property is not strictly enforced, and does not cover all aspects of the object's
        state. It does however provide a general guide for whether or not a modification is able
        to be translated from :mod:`glycopeptidepy`'s internal representation into something that
        another program *should* be able to understand.

        Returns
        -------
        bool
        """
        return True


cdef class NeutralLossBase(object):
    def __eq__(self, other):
        if other is None:
            return False
        return self.label == (<NeutralLossBase>other).label and self.mass == (<NeutralLossBase>other).mass

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.label)


cdef EnumValue ModificationCategory_glycosylation = ModificationCategory.glycosylation


cdef class GlycosylationBase(object):
    cdef bint get_is_composition(self):
        return self._is_composition

    cdef void set_is_composition(self, bint value):
        self._is_composition = value

    cdef bint get_is_core(self):
        return self._is_core

    cdef void set_is_core(self, bint value):
        self._is_core = value

    cdef EnumValue get_glycosylation_type(self):
        return self._glycosylation_type

    cdef void set_glycosylation_type(self, EnumValue glycosylation_type):
        self._glycosylation_type = glycosylation_type

    cpdef bint is_tracked_for(self, object category):
        return ModificationCategory_glycosylation == category

    @property
    def is_core(self):
        return self.get_is_core()

    @is_core.setter
    def is_core(self, bint value):
        self.set_is_core(value)

    @property
    def is_composition(self):
        return self.get_is_composition()

    @is_composition.setter
    def is_composition(self, bint value):
        self.set_is_composition(value)

    @property
    def glycosylation_type(self):
        return self.get_glycosylation_type()

    @glycosylation_type.setter
    def glycosylation_type(self, value):
        self.set_glycosylation_type(value)


@cython.binding(True)
cpdef bint is_tracked_for_glycosylation(self, object category):
    """Determine if this :class:`ModificationBase` is tracked by a particular
    behavioral pattern associated with a :class:`~.ModificationCategory`.

    This relationship is distinct from :meth:`is_a` which merely observes that
    the semantic relationship holds, not that any actual behavior is available.

    Parameters
    ----------
    category : :class:`~.ModificationCategory`
        The category to check

    Returns
    -------
    bool
    """
    return category == ModificationCategory_glycosylation


from cpython cimport PyObject

cdef struct count_table_bin_cell:
    long value
    PyObject* key


cdef struct count_table_bin:
    count_table_bin_cell* cells
    size_t size
    size_t used


cdef struct count_table:
    count_table_bin* bins
    size_t size
    size_t count


cdef int initialize_count_table_bin(count_table_bin* bin, size_t size)
cdef void free_count_table_bin(count_table_bin* bin)
cdef int count_table_bin_append(count_table_bin* bin, PyObject* key, long value)
cdef int count_table_bin_find(count_table_bin* bin, PyObject* query, Py_ssize_t* cell_index)

cdef count_table* make_count_table(size_t table_size, size_t bin_size)
cdef void free_count_table(count_table* table)
cdef int count_table_find_bin(count_table* table, PyObject* query, Py_ssize_t* bin_index) except 1
cdef int count_table_put(count_table* table, PyObject* key, long value) except 1
cdef int count_table_del(count_table* table, PyObject* key, long* value) except 1
cdef int count_table_get(count_table* table, PyObject* key, long* value) except 1
cdef int count_table_increment(count_table* table, PyObject* key, long value)
cdef int count_table_decrement(count_table* table, PyObject* key, long value)

cdef Py_ssize_t count_table_count(count_table* table)
cdef long count_table_total(count_table* table)

cdef list count_table_keys(count_table* table)
cdef list count_table_values(count_table* table)
cdef list count_table_items(count_table* table)

cdef void count_table_add(count_table* table_a, count_table* table_b)
cdef void count_table_subtract(count_table* table_a, count_table* table_b)
cdef void count_table_scale(count_table* table, long value)

cdef void count_table_update(count_table* table_a, count_table* table_b)
cdef void count_table_clear(count_table* table)
cdef count_table* count_table_copy(count_table* table_a)

cdef bint count_table_equals(count_table* table_a, count_table* table_b)


cdef class CountTableIterator(object):
    cdef :
        count_table* table
        size_t bin_index
        size_t cell_index
        PyObject* next_key
        long next_value

    @staticmethod
    cdef CountTableIterator _create(CountTable table)

    cdef void advance(self)
    cdef bint has_more(self)
    cdef int _locate_next_value(self, PyObject** key, long* value)
    cdef int get_next_value(self, PyObject** key, long* value)


cdef class CountTable(object):
    cdef:
        count_table* table

    @staticmethod
    cdef CountTable _create()

    @staticmethod
    cdef CountTable _create_from(CountTable table)
    cdef int _initialize_table(self) except 1

    cdef void _update_from_dict(self, dict d)
    cdef void _update_from_count_table(self, CountTable other)

    cdef void _add_from(self, CountTable other)
    cdef void _subtract_from(self, CountTable other)
    cdef int _add_from_dict(self, dict other) except 1
    cdef int _subtract_from_dict(self, dict other) except 1
    cdef void _scale_by(self, long value)

    cdef bint equal_to(self, CountTable other)

    cpdef dict _to_dict(self)

    cpdef update(self, obj)
    cpdef CountTable copy(self)
    cpdef list items(self)
    cpdef list values(self)
    cpdef list keys(self)
    cpdef clear(self)
    cpdef setdefault(self, key, value)
    cpdef long pop(self, object key, object default=?) except *
    cpdef get(self, object key, object default=?)

    cdef void increment(self, object key, long value)
    cdef void decrement(self, object key, long value)
    cdef long getitem(self, object key) except *
    cdef int setitem(self, object key, long value) except 1
    cdef long delitem(self, object key) except *
    cpdef long total(self)


cimport cython

from glypy.composition.ccomposition cimport CComposition

from glycopeptidepy._c.count_table cimport CountTable
from glycopeptidepy._c.structure.fragment cimport IonSeriesBase, PeptideFragment, ChemicalShiftBase
from glycopeptidepy._c.structure.sequence_methods cimport _PeptideSequenceCore
from glycopeptidepy._c.structure.base cimport ModificationBase, AminoAcidResidueBase, SequencePosition
from glycopeptidepy._c.structure.modification.modification cimport ModificationInstanceBase

from glycopeptidepy._c.structure.fragmentation_strategy.base cimport FragmentationStrategyBase

from glycopeptidepy.structure.modification import (
    Modification,
    NGlycanCoreGlycosylation,
    OGlycanCoreGlycosylation,
    GlycosaminoglycanLinkerGlycosylation)

from glycopeptidepy.structure.glycan import (GlycosylationType)


cdef:
    dict hcd_modifications_of_interest
    dict hcd_modification_compositions
    dict hcd_modifications_of_interest_to_variants_cache


cdef class PeptideFragmentationStrategyBase(FragmentationStrategyBase):
    cdef:
        public IonSeriesBase series
        public dict chemical_shift_rules
        public int max_chemical_shifts
        public int direction

        public bint include_neutral_losses

        public double running_mass
        public double running_delta_mass
        public CComposition running_composition
        public long index
        public long size
        public CountTable modification_index
        public dict glycosylation_manager
        public CountTable amino_acids_counter

    cpdef _initialize_fields(self)
    cpdef _initialize_start_terminal(self)
    cpdef list _get_viable_chemical_shift_combinations(self)

    cpdef CComposition composition_of(self, SequencePosition position)
    cpdef long name_index_of(self)
    cpdef list flanking_residues(self)
    cpdef track_glycosylation(self, long index, glycosylation)

    cpdef bint has_more(self)
    cpdef list _build_fragments(self)
    cpdef list partial_loss(self, PeptideFragment fragment)
    cpdef object step(self)

    cpdef _update_state(self)
    cpdef reset(self)


@cython.freelist(1000)
cdef class ModificationConfiguration(object):
    cdef:
        # the modifications of interest to HCDFragmentationStrategy
        public CountTable modifications_of_interest
        # the modifications not of interest to HCDFragmentationStrategy
        public CountTable other_modifications
        # the composition shift associated with the interesting modifications
        public CComposition delta_composition
        # the set of all modifications, both interesting and not interesting, used as
        # an identity key for this configuration
        public CountTable modification_set
        # the mass shift associated with the other modifications
        public double other_modifications_mass

    @staticmethod
    cdef ModificationConfiguration _create(CountTable modifications_of_interest, CountTable other_modifications,
                                           CComposition delta_composition, CountTable modification_set,
                                           double other_modifications_mass)

    cdef bint equal_to(self, ModificationConfiguration other)


cdef class HCDFragmentationStrategy(PeptideFragmentationStrategyBase):
    cdef:
        public ModificationConfiguration _last_modification_set
        public list _last_modification_variants

    cpdef _get_core_for(self, ModificationInstanceBase glycosylation)
    cpdef ModificationConfiguration _get_modifications_of_interest(self, PeptideFragment fragment)
    cpdef _replace_cores(self, CountTable modifications_of_interest)
    cpdef list _generate_modification_variants(self, CountTable interesting_modifications, CountTable other_modifications)

    cpdef double[::1] mass_series(self)



from typing import Dict, List, Optional, Type, Union
from six import string_types as basestring

from glycopeptidepy.structure.fragmentation_strategy.peptide import PeptideFragmentationStrategyBase

from ..fragmentation_strategy import HCDFragmentationStrategy
from ..fragment import ChemicalShift, IonSeries
from ..base import ModificationBase, SequencePosition
from ..modification import SequenceLocation
from ..residue import AminoAcidResidue

from .base import _PeptideSequenceCore
from .glycosylated_sequence import GlycosylatedSequenceMixin
from .mutable_sequence import MutableSequenceMixin


b_series = IonSeries.b
y_series = IonSeries.y
stub_glycopeptide_series = IonSeries.stub_glycopeptide


class PeptideSequence(_PeptideSequenceCore, GlycosylatedSequenceMixin, MutableSequenceMixin):
    @classmethod
    def from_iterable(cls, iterable, glycan_composition=None, n_term=None, c_term=None, text=None, **kwargs):
        """Construct a :class:`PeptideSequence` instance from an arbitrary iterable of (amino acid, [modifications...])
        pairs.

        Parameters
        ----------
        iterable: :class:`~.Iterable`
            An arbitrary :class:`~.Iterable` of two-element pairs where the first element denotes
            the amino acid at that position while the second is a list of modifications. These elements
            may be denoted by actual objects of the appropriate type or as strings (so long as all are are
            typed object or strings) and will be coerced accordingly.
        glycan_composition: :class:`str` or :class:`~.GlycanComposition`
            An optional glycan composition given either as a string or a typed object. Will be coerced
            as needed.
        n_term: :class:`~.TerminalGroup`
            The amino terminal group
        c_term: :class:`~.TerminalGroup`
            The carboxyl terminal group
        text: :class:`bool`
            Whether or not to enforce text parsing. Defaults to :const:`None` which results in automatic
            detection.

        Returns
        -------
        :class:`PeptideSequence`
        """
        seq = cls()
        if not isinstance(iterable, list):
            iterable = list(iterable)
        if text is None:
            try:
                if isinstance(iterable[0][0], basestring):
                    text = True
            except IndexError:
                text = True
        if text:
            seq._init_from_parsed_string(iterable, glycan_composition, n_term, c_term)
        else:
            try:
                if not isinstance(iterable[0], SequencePosition):
                    # Coerce all pairs in the input to :class:`SequencePosition` because :meth:`_init_from_components`
                    # assumes that its input is a strictly typed list.
                    iterable = list(map(SequencePosition, iterable))
            except (ValueError, AttributeError, TypeError):
                iterable = list(map(SequencePosition, iterable))
            seq._init_from_components(
                iterable,
                glycan_composition, n_term, c_term, **kwargs)
        return seq

    def modified_residues(self):
        """Find all the modified residue positions in the sequence

        Returns
        -------
        list of :class:`~.SequencePosition`
        """
        return [(i, position) for i, position in enumerate(self) if position.modifications]

    def has_modification(self, position: Union[int, SequenceLocation],
                         modification_type: Union[str, ModificationBase]) -> bool:
        '''Check if a modification is present on the sequence

        Parameters
        ----------
        position: int
            The position in the sequence to check for the modification
        modification_type: str or Modification
            The modification to check for.

        Returns
        -------
        bool:
            Whether or not `modification_typ` is present at `position`
        '''
        if position is SequenceLocation.n_term:
            return self.n_term == modification_type
        elif position is SequenceLocation.c_term:
            return self.c_term == modification_type
        return self.sequence[position].has_modification(modification_type)


    def strip_modifications(self) -> 'PeptideSequence':
        """Return a copy of this sequence with all modifications removed.

        Returns
        -------
        :class:`PeptideSequence`
        """
        parts = [[pos.amino_acid, []] for pos in self]
        return self.from_iterable(parts)

    def subsequence(self, slice_obj: slice) -> 'PeptideSequence':
        sub = self[slice_obj]
        subseq = self.from_iterable(sub)
        if slice_obj.start == 0:
            subseq.n_term = self.n_term
        if slice_obj.stop == len(self):
            subseq.c_term = self.c_term
        return subseq

    def reverse(self):
        slc = slice(None, None, -1)
        seq = self.subsequence(slc)
        if seq.glycan_composition != self.glycan_composition:
            if seq.glycan_composition:
                raise ValueError("Cannot recalculate partially ambiguous glycan composition")
            else:
                seq.glycan = self.glycan_composition.clone()
        return seq

    def get_fragments(self,
                      kind: Union[str, IonSeries],
                      chemical_shifts: Dict[AminoAcidResidue, List[ChemicalShift]]=None,
                      strategy: Optional[Type[PeptideFragmentationStrategyBase]]=None,
                      include_neutral_losses: bool=False,
                      **kwargs) -> PeptideFragmentationStrategyBase:
        """Generate fragments from this structure according to the strategy specified
        by ``strategy``, returning an iterator over the sequence of theoretical fragments.

        There may be multiple fragments for a single position, resulting in the iterator
        yielding lists of fragment objects per step.

        Parameters
        ----------
        kind : :class:`~.IonSeries` or :class:`str`
            The name of the ion series to produce fragments from, either as a string or the
            :class:`~.IonSeries` object to use
        chemical_shifts : dict, optional
            A :class:`~.Mapping` between :class:`~.AminoAcidResidue` and a listof acceptable chemical shifts to apply
            to the produced fragments containing that :class:`~.AminoAcidResidue`, to be applied combinatorially.
        strategy : :class:`~.FragmentationStrategyBase` type, optional
            The strategy type to employ when producing fragments. Defaults to :class:`~.HCDFragmentationStrategy`.
        include_neutral_losses: :class:`bool`
            Whether to include generic neutral losses (-NH3 and -H2O) on all fragments
        **kwargs
            Passed to ``strategy``

        Returns
        -------
        :class:`~.FragmentationStrategyBase`
            The fragmentation iterator
        """
        if strategy is None:
            strategy = HCDFragmentationStrategy
        losses = kwargs.pop('neutral_losses', {})
        if losses and not chemical_shifts:
            chemical_shifts = losses
        if kind == stub_glycopeptide_series:
            return ([frag] for frag in self.stub_fragments(True))
        else:
            return strategy(
                self, kind, chemical_shifts=chemical_shifts,
                include_neutral_losses=include_neutral_losses, **kwargs)


def list_to_sequence(seq_list, wrap=True, peptide_type=PeptideSequence):
    flat_chunks = []
    for chunk in seq_list:
        if(isinstance(chunk[0], list)):
            flat_chunks.extend(chunk)
        else:
            flat_chunks.append(chunk)
    seq = peptide_type.from_iterable(flat_chunks) if wrap else flat_chunks
    return seq


Sequence = PeptideSequence
# parse = Sequence


class NamedSequence(PeptideSequence):

    def __init__(self, name=None, sequence=None, parser_function=None, **kwargs):
        super(NamedSequence, self).__init__(
            sequence, parser_function, **kwargs)
        self.name = name

    def clone(self):
        dup = super(NamedSequence, self).clone()
        dup.name = self.name
        return dup

    def __repr__(self):
        string = super(NamedSequence, self).__str__()
        name = str(self.name)
        if name.startswith(">"):
            name = name[1:]
        return ">%s\n%s" % (name, string)


class AnnotatedSequence(NamedSequence):

    def __init__(self, name=None, sequence=None, parser_function=None, annotations=None,
                 **kwargs):
        super(AnnotatedSequence, self).__init__(
            name, sequence, parser_function=parser_function, **kwargs)
        self.annotations = self._prepare_annotations(annotations)

    @staticmethod
    def _prepare_annotations(annotation_data):
        if annotation_data:
            return dict(annotation_data)
        else:
            return dict()

    def clone(self):
        dup = super(AnnotatedSequence, self).clone()
        dup.annotations = self._prepare_annotations(self.annotations)
        return dup


class ProteinSequence(AnnotatedSequence):
    pass



import unittest

from glycopeptidepy.structure import sequence, modification, residue
from glycopeptidepy.structure import Composition

from glypy.io import glycoct

n_glycan_rule = '#:glycoct:RES 1b:b-dglc-HEX-1:5 2s:n-acetyl 3b:b-dglc-HEX-1:5 \
4s:n-acetyl 5b:b-dman-HEX-1:5 6b:a-dman-HEX-1:5 7b:a-dman-HEX-1:5 LIN \
1:1d(2+1)2n 2:1o(4+1)3d 3:3d(2+1)4n 4:3o(4+1)5d 5:5o(3+1)6d 6:5o(6+1)7d'


class TestModifcationTarget(unittest.TestCase):
    def test_n_term_special_flag(self):
        mt = modification.ModificationTable()
        rule = mt["Acetylation"]
        peptide = sequence.parse("QVPQLQSQ")
        sites = rule.find_valid_sites(peptide)
        self.assertTrue(sites[0] is modification.SequenceLocation.n_term)

        rule = mt['Gln->pyro-Glu']
        sites = rule.find_valid_sites(peptide)
        self.assertEqual(sites[0], 0)

        deamidated = mt["Deamidated"]
        sites = deamidated.find_valid_sites(peptide)
        self.assertEqual(sites, [0, 3, 5, 7])

    def test_parse_target(self):
        spaced = 'Q @ N-term'
        part_spaced = 'Q @N-Term'
        part_spaced2 = 'Q@ n-term'
        no_space = 'Q@N-term'
        spaced_parsed = modification.extract_targets_from_string(spaced)
        self.assertEqual(spaced_parsed, modification.extract_targets_from_string(part_spaced))
        self.assertEqual(spaced_parsed, modification.extract_targets_from_string(part_spaced2))
        self.assertEqual(spaced_parsed, modification.extract_targets_from_string(no_space))

    def test_serialize(self):
        spaced = 'Q @ N-term'
        spaced_parsed = modification.extract_targets_from_string(spaced)
        assert spaced == spaced_parsed.serialize()

    def test__len__(self):
        spaced = 'Q @ N-term'
        spaced_parsed = modification.extract_targets_from_string(spaced)
        assert len(spaced_parsed) == 1


class TestGlycosylationRule(unittest.TestCase):
    def test_parse(self):
        rule = modification.Glycosylation.try_parse(n_glycan_rule)
        base_mass = glycoct.loads("\
            RES 1b:b-dglc-HEX-1:5 2s:n-acetyl \
            3b:b-dglc-HEX-1:5 4s:n-acetyl \
            5b:b-dman-HEX-1:5 6b:a-dman-HEX-1:5 \
            7b:a-dman-HEX-1:5 LIN 1:1d(2+1)2n \
            2:1o(4+1)3d 3:3d(2+1)4n 4:3o(4+1)5d \
            5:5o(3+1)6d 6:5o(6+1)7d").mass()
        dehydrated_mass = base_mass - Composition("H2O").mass
        self.assertAlmostEqual(rule.mass, dehydrated_mass, 3)


class TestModificationTable(unittest.TestCase):
    def test_restricted_table(self):
        table = modification.RestrictedModificationTable(
            constant_modifications=['Carbamidomethyl (C)'],
            variable_modifications=['Deamidated (N)'])

        total = len(table.rules())
        given = len(table.other_modifications)
        assert total - given == 2

    def test_by_unimod(self):
        mod = modification.Modification("UNIMOD:4")
        assert mod.name == "Carbamidomethyl"


class TestModificationRule(unittest.TestCase):
    def test_targets(self):
        mod = modification.Modification("Deamidation")
        assert len(mod.rule.targets) == 5
        assert len(mod.rule.n_term_targets) == 1
        assert len(mod.rule.c_term_targets) == 0

    def test_subtraction(self):
        mod = modification.Modification("Deamidation")
        assert len((mod.rule - mod.rule).targets) == 0

    def test_serialize(self):
        mod = modification.Modification("Deamidation")
        assert mod.rule.serialize() == mod.rule.name


class TestAminoAcidSubstitution(unittest.TestCase):
    def test_parse(self):
        rule_name = "@Asn->Pro"
        rule = modification.AminoAcidSubstitution.try_parse(rule_name)
        assert rule is not None
        assert rule.serialize() == rule_name


class AnonymousModificationRule(unittest.TestCase):
    def test_parse(self):
        rule_name = "@Name-1.0"
        rule = modification.AnonymousModificationRule.try_parse(rule_name)
        assert rule.name == "Name"
        assert rule.mass == 1.0
        assert rule.serialize() == rule_name
        assert not rule.is_standard
        assert rule == rule.clone()


if __name__ == '__main__':
    unittest.main()



import unittest

from glycopeptidepy.structure import sequence, modification, residue, sequence_composition


class TestSequenceComposition(unittest.TestCase):
    def test_parse(self):
        composition_string = '{N:2; K:1; E:2; Y:1}'
        composition = sequence_composition.SequenceComposition.parse(composition_string)
        seq = sequence.PeptideSequence("NEEYNK")
        self.assertAlmostEqual(seq.mass, composition.mass, 3)

    def test_serialize(self):
        composition = sequence_composition.SequenceComposition()
        composition['N'] = 2
        composition['K'] = 1
        composition['E'] = 2
        composition['Y'] = 1
        assert composition.serialize() == '{N:2; K:1; E:2; Y:1}'

    def test_clone(self):
        composition_string = '{N:2; K:1; E:2; Y:1}'
        composition = sequence_composition.SequenceComposition.parse(composition_string)
        assert composition == composition.clone()

    def test_modified(self):
        composition_string = '{N:1; N(N-Glycosylation):1; K:1; E:2; Y:1}'
        composition = sequence_composition.SequenceComposition.parse(composition_string)
        self.assertAlmostEqual(composition.mass, 998.4192, 3)


if __name__ == '__main__':
    unittest.main()



from .constants import constants

from . import composition
from .composition import Composition, formula, ChemicalCompositionError

from . import base
from .base import (
    MoleculeBase, PeptideSequenceBase, ModificationBase,
    ResidueBase, SequencePosition)

from . import parser
from .parser import (sequence_tokenizer, strip_modifications)

from . import residue
from .residue import (
    AminoAcidResidue, register_degenerate, register_residue)

from .terminal_group import TerminalGroup

from . import sequence
from .sequence import (
    PeptideSequence, NamedSequence, ProteinSequence,
    AnnotatedSequence,
    find_glycosaminoglycan_sequons, find_n_glycosylation_sequons,
    find_o_glycosylation_sequons, )

from . import modification
from .modification import (
    Modification, ModificationRule, ModificationTarget,
    ModificationTable, RestrictedModificationTable,
    AnonymousModificationRule, AminoAcidSubstitution,
    ModificationCategory, SequenceLocation, Glycosylation,
    CoreGlycosylation, NGlycanCoreGlycosylation, OGlycanCoreGlycosylation,
    GlycosaminoglycanLinkerGlycosylation, OGlcNAcylation, GlycanFragment,
    ModificationStringParseError, ModificationNameResolutionError)

from . import fragment
from .fragment import (
    ChemicalShift, FragmentBase, PeptideFragment,
    SimpleFragment, StubFragment, IonSeries)

from . import fragmentation_strategy
from .fragmentation_strategy import (
    FragmentationStrategyBase, CADFragmentationStrategy, HCDFragmentationStrategy,
    EXDFragmentationStrategy, StubGlycopeptideStrategy)

from . import glycan
from .glycan import (
    TypedGlycan, TypedGlycanComposition, GlycosylationType,
    GlycanCompositionProxy, GlycosylationManager, )


__all__ = [
    "constants",

    "base",
    "MoleculeBase", "PeptideSequenceBase", "ModificationBase",
    "ResidueBase", "SequencePosition",

    "parser",
    "sequence_tokenizer", "strip_modifications",

    "residue",
    "AminoAcidResidue", "register_degenerate", "register_residue",

    "sequence",
    "PeptideSequence", "NamedSequence", "ProteinSequence",
    "AnnotatedSequence", "TerminalGroup", "find_glycosaminoglycan_sequons",
    "find_o_glycosylation_sequons", "find_n_glycosylation_sequons",

    "modification",
    "Modification", "ModificationRule", "ModificationTarget",
    "ModificationTable", "RestrictedModificationTable",
    "AnonymousModificationRule", "AminoAcidSubstitution",
    "ModificationCategory", "SequenceLocation", "Glycosylation",
    "CoreGlycosylation", "NGlycanCoreGlycosylation", "OGlycanCoreGlycosylation",
    "GlycosaminoglycanLinkerGlycosylation", "OGlcNAcylation", "GlycanFragment",
    "ModificationStringParseError", "ModificationNameResolutionError",

    "composition", "Composition", "formula", "ChemicalCompositionError",

    "fragment",
    "ChemicalShift", "FragmentBase", "PeptideFragment",
    "SimpleFragment", "StubFragment", "IonSeries",

    "fragmentation_strategy",
    "FragmentationStrategyBase", "CADFragmentationStrategy", "HCDFragmentationStrategy",
    "EXDFragmentationStrategy", "StubGlycopeptideStrategy",

    "glycan",
    "TypedGlycan", "TypedGlycanComposition", "GlycosylationType",
    "GlycanCompositionProxy", "GlycosylationManager",

]



import unittest
import requests
from glycopeptidepy.io import uniprot


try:
    is_online = True
    response = requests.get("http://www.uniprot.org/")
    response.raise_for_status()
except Exception:
    is_online = False


def skip_not_online(fn):
    if not is_online:
        return unittest.skip("Not able to reach host")(fn)
    else:
        return fn


@skip_not_online
class UniProtClientTest(unittest.TestCase):
    def test_get(self):
        prot = uniprot.get("P13611")
        self.assertIn("P13611", prot.accessions)

    def test_search(self):
        result = uniprot.search('Aggrecan')
        for res in result:
            if res['Entry'] == 'P16112':
                break
        else:
            raise AssertionError("Did not find Human Aggrecan in the search results")



cimport cython
from glypy.composition.ccomposition cimport CComposition


cdef class PeptideSequenceBase(object):
    cdef:
        public list sequence
        public double _mass
        public object _glycosylation_manager
        public TerminalGroup _n_term
        public TerminalGroup _c_term

        public CComposition _total_composition
        public CComposition _peptide_composition

    cdef SequencePosition get(self, ssize_t i)

    cpdef _invalidate(self)


cdef class TerminalGroup(object):
    cdef:
        public CComposition base_composition
        public double mass
        public ModificationBase _modification

    cdef ModificationBase get_modification(self)
    cdef CComposition get_composition(self)
    cpdef TerminalGroup  modify(self, ModificationBase modification)
    cdef void set_modification(self, ModificationBase)
    cdef bint equal_to(self, object other)
    cpdef str serialize(self)


cdef class AminoAcidResidueBase(object):
    cdef:
        public basestring name
        public basestring symbol
        public double mass
        public CComposition composition
        public object neutral_loss
        public Py_hash_t _hash

    @staticmethod
    cdef AminoAcidResidueBase _create(str name, str symbol, double mass, CComposition composition)

    cdef bint equal_to(self, AminoAcidResidueBase other)


cdef class ModificationBase(object):
    cdef:
        public basestring name
        public double mass
        public CComposition composition

    cpdef bint is_a(self, object category)

    cpdef basestring serialize(self)

    cpdef bint is_tracked_for(self, object category)


ctypedef fused modification_or_string:
    str
    bytes
    ModificationBase


cdef class SequencePosition(object):
    cdef:
        public AminoAcidResidueBase amino_acid
        public list modifications

    @staticmethod
    cdef SequencePosition _create(AminoAcidResidueBase amino_acid, list modifications)

    cdef inline double get_mass(self)

    cpdef bint has_modification(self, modification)
    cpdef bint is_modified(self)
    cpdef drop_modification(self, modification)
    cpdef add_modification(self, modification)

    cdef inline size_t get_modification_count(self)
    cdef inline ModificationBase get_modification(self, size_t i)


from glypy.utils import Enum


class SequenceLocation(Enum):
    anywhere = None
    n_term = -1
    c_term = -2
    protein_n_term = -3
    protein_c_term = -4


SequenceLocation.n_term.add_name("N-term")
SequenceLocation.n_term.add_name("N-Term")
SequenceLocation.n_term.add_name("n-term")
SequenceLocation.n_term.add_name("N_term")
SequenceLocation.c_term.add_name("C-term")
SequenceLocation.c_term.add_name("C-Term")
SequenceLocation.c_term.add_name("c-term")
SequenceLocation.c_term.add_name("C_term")
SequenceLocation.anywhere.add_name("Anywhere")
SequenceLocation.protein_n_term.add_name("Protein N-term")
SequenceLocation.protein_c_term.add_name("Protein C-term")
SequenceLocation.protein_n_term.add_name("Protein N-Term")
SequenceLocation.protein_c_term.add_name("Protein C-Term")


class ModificationCategory(Enum):
    unknown = None
    glycosylation = 1
    artefact = 2
    substitution = 3
    chemical_derivative = 4
    non_standard_residue = 5
    isotopic_label = 6
    post_translational = 7
    other = 8
    multiple = 10
    pre_translational = 11
    co_translational = 12
    synthetic_peptide_protect = 13
    cross_link = 14
    cid_cleavable_cross_link = 15
    other_cleavable_cross_link = 16

ModificationCategory.substitution.add_name("AA substitution")
ModificationCategory.glycosylation.add_name("other_glycosylation")
ModificationCategory.glycosylation.add_name("N-linked glycosylation")
ModificationCategory.glycosylation.add_name("O-linked glycosylation")
ModificationCategory.glycosylation.add_name("Other glycosylation")
ModificationCategory.chemical_derivative.add_name("Chemical derivative")
ModificationCategory.post_translational.add_name("Post-translational")
ModificationCategory.multiple.add_name("Multiple")
ModificationCategory.artefact.add_name("Artefact")
ModificationCategory.isotopic_label.add_name("Isotopic label")
ModificationCategory.other.add_name("Other")
ModificationCategory.pre_translational.add_name("Pre-translational")
ModificationCategory.non_standard_residue.add_name("Non-standard residue")
ModificationCategory.co_translational.add_name("Co-translational")
ModificationCategory.synthetic_peptide_protect.add_name("Synth. pep. protect. gp.")
ModificationCategory.cross_link.add_name("Cross-link")
ModificationCategory.cid_cleavable_cross_link.add_name("CID cleavable cross-link")
ModificationCategory.other_cleavable_cross_link.add_name("Other cleavable cross-link")


import re
import string
import warnings

from collections import defaultdict

from .descriptors import SequenceLocation, ModificationCategory

from ..residue import AminoAcidResidue


target_string_pattern = re.compile(
    r'''(?P<amino_acid>[A-Z]+)?
        (@
            (?P<n_term>[Nn][-_][tT]erm)|
            (?P<c_term>[Cc][-_][tT]erm)
        )?''',
    re.VERBOSE)
title_cleaner = re.compile(
    r'(?P<name>.*)\s\((?P<target>.+)\)$')

is_mass_delta = re.compile(r"^Delta:")


def extract_targets_from_string(target_string):
    position_only = re.search(r"^((?:Protein )?[NnCc][-_][tT]erm)", target_string)
    if position_only:
        return ModificationTarget(None, SequenceLocation[position_only.groups()[0]])
    else:
        state = 'aa'
        amino_acids = []
        position = []

        protein_terminal_only = False
        i = 0
        n = len(target_string)
        while i < n:
            c = target_string[i]

            if c == " ":
                if state == 'aa':
                    state = "aa-space"
                elif state == 'position-begin':
                    pass
                elif state == 'protein':
                    state = 'position-begin'
                    protein_terminal_only = True
                else:
                    raise ValueError(target_string)
            elif c == "@":
                if state == 'aa':
                    state = 'position-begin'
                elif state == 'aa-space':
                    state = 'position-begin'
                else:
                    raise ValueError("Found @ not between amino acids and position")
            elif c in string.ascii_uppercase:
                if state == 'aa':
                    amino_acids.append(c)
                elif state == 'position-begin':
                    if c in "NC":
                        position.append(c)
                        state = 'position'
                    elif c == "P":
                        state = 'protein'
                elif state == 'position':
                    if c == 'T':
                        position.append(c)
                    else:
                        raise ValueError("Invalid position specification")
            elif c in string.ascii_lowercase:
                if state == 'aa':
                    amino_acids.append(c.upper())
                elif state == 'position-begin':
                    if c in 'nc':
                        position.append(c)
                        state = 'position'
                    if c == 'p':
                        state = 'protein'
                elif state == "protein":
                    if c in 'protein':
                        pass
                    else:
                        raise ValueError("Invalid terminal speicification")
                elif state == 'position':
                    position.append(c)
            elif c in '-_':
                if state == 'position':
                    position.append(c)
                else:
                    raise ValueError("Found - not in position specification")
            i += 1

        position = ''.join(position)
        position = position.lower().replace("-", "_").strip()

        amino_acids = list(map(AminoAcidResidue, amino_acids))
        if position == "":
            position = SequenceLocation.anywhere
        elif position == "n_term":
            if not protein_terminal_only:
                position = SequenceLocation.n_term
            else:
                position = SequenceLocation.protein_n_term
        elif position == "c_term":
            if not protein_terminal_only:
                position = SequenceLocation.c_term
            else:
                position = SequenceLocation.protein_c_term
        else:
            raise ValueError("Unrecognized position specification {}".format(position))
        return ModificationTarget(amino_acids, position)


def extract_targets_from_rule_string(name):
    return extract_targets_from_string(title_cleaner.search(name).groupdict()["target"])


def get_position_modifier_rules_dict(sequence, protein_n_term=False, protein_c_term=False):
    '''Labels the start and end indices of the sequence

    A convenience dictionary initializer for ease-of-access in
    :class:`ModificationTarget` site validation functions

    Parameters
    ----------
    sequence : PeptideSequence
        The sequence for which to project labels
    protein_n_term : bool
        Whether the sequence includes a protein N-terminal
    protein_c_term : bool
        Whether the sequence includes a protein C-terminal

    Return
    ------
    defaultdict
    '''
    if protein_n_term:
        n_term = (SequenceLocation.n_term, SequenceLocation.protein_n_term)
    else:
        n_term = (SequenceLocation.n_term, )
    if protein_c_term:
        c_term = (SequenceLocation.c_term, SequenceLocation.protein_c_term)
    else:
        c_term = (SequenceLocation.c_term, )
    return defaultdict(lambda: SequenceLocation.anywhere, {
        0: n_term,
        (len(sequence) - 1): c_term
    })


class ModificationTarget(object):
    '''Specifies the range of targets that a given modification can be found at.

    A single instance of :class:`ModificationTarget` describes one or more position/residue
    rules for a single type of modification, bundled together underneath a single :class:`ModificationRule`
    instance. :class:`ModificationTarget` may be extracted from Unimod data dictionaries using the
    :meth:`from_unimod_specificity` class method.

    Attributes
    ----------
    amino_acid_targets: list
        The list of amino acid residues that this rule
        permits.
    position_modifier: SequenceLocation
        The sequence-position specific filter
        this rule enforces.
    '''

    @classmethod
    def from_unimod_specificity(cls, specificity):
        amino_acid = (specificity["site"])
        if amino_acid in (["N-term", "C-term", None]):
            amino_acid = None
        else:
            amino_acid = {AminoAcidResidue(amino_acid)}
        position_modifier = specificity["position"]
        nterm = ("Any N-term", "N-term")
        prot_nterm = ("Protein N-term", )
        cterm = ("Any C-term", "C-term")
        prot_cterm = ("Protein C-term", )
        site = specificity["site"]
        classification = ModificationCategory[specificity.get("classification")]
        if position_modifier == "Anywhere":
            position_modifier = SequenceLocation.anywhere
        elif position_modifier in nterm or site in nterm:
            position_modifier = SequenceLocation.n_term
        elif position_modifier in cterm or site in cterm:
            position_modifier = SequenceLocation.c_term
        elif position_modifier in prot_nterm:
            position_modifier = SequenceLocation.protein_n_term
        elif position_modifier in prot_cterm:
            position_modifier = SequenceLocation.protein_c_term
        else:
            warnings.warn("Could not find a location for %r" % (position_modifier,))

        return cls(amino_acid, position_modifier, [classification])

    def __init__(self, site_targets=None, position_modifier=None, classification=None):
        if classification is None:
            classification = []
        self.amino_acid_targets = frozenset(map(AminoAcidResidue, site_targets)) if site_targets is not None else None
        self.position_modifier = SequenceLocation[position_modifier]
        self.classification = classification

    def valid_site(self, amino_acid=None, position_modifier=SequenceLocation.anywhere):
        '''Return if a given residue at a sequence position (N-term, C-term, None)
        is valid

        Arguments
        ---------
        amino_acid: AminoAcidResidue
            The amino acid to test
        position_modifier: SequenceLocation
            The position in the sequence to test

        Returns
        -------
        valid: bool
            Whether the match is valid
        location_specification: SequenceLocation
            If the site is valid, this will be the type of location matched,
            otherwise it will default to SequenceLocation.anywhere
        '''
        valid = False
        if not isinstance(position_modifier, (tuple, list)):
            position_modifiers = (position_modifier, )
        else:
            position_modifiers = position_modifier

        for position_modifier in position_modifiers:
            # Validate amino acid target
            valid = (self.amino_acid_targets is None) or (
                amino_acid in self.amino_acid_targets)
            valid = valid and ((self.position_modifier == SequenceLocation.anywhere) or
                            (position_modifier == self.position_modifier))

            # If the rule includes a position modifier other than anywhere
            if valid and (position_modifier != SequenceLocation.anywhere) and (
                    self.position_modifier != SequenceLocation.anywhere):
                if position_modifier == self.position_modifier:
                    if self.amino_acid_targets is None:
                        return valid, position_modifier
                    else:
                        return valid, SequenceLocation.anywhere
                else:
                    valid = False
        return valid, SequenceLocation.anywhere

    def __repr__(self):
        amino_acid_components = "{%s}" % ', '.join(map(str, self.amino_acid_targets or ()))
        rep = "{amino_acid_targets}@{position_modifier}".format(
            amino_acid_targets=amino_acid_components,
            position_modifier=self.position_modifier)
        return rep

    def __eq__(self, other):
        if self.amino_acid_targets != other.amino_acid_targets:
            return False
        elif self.position_modifier != other.position_modifier:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.amino_acid_targets)

    def __len__(self):
        position_modifier_count = 0
        amino_acid_targets_count = 0
        if self.amino_acid_targets is not None:
            amino_acid_targets_count = len(self.amino_acid_targets)
        return amino_acid_targets_count + position_modifier_count

    def serialize(self):
        parts = []
        if self.amino_acid_targets is not None:
            parts.append(''.join(aa.symbol for aa in self.amino_acid_targets))
            if self.position_modifier == SequenceLocation.anywhere:
                pass
            else:
                parts.append("@")
        if self.position_modifier == SequenceLocation.n_term:
            parts.append("N-term")
        elif self.position_modifier == SequenceLocation.protein_n_term:
            parts.append("Protein N-term")
        elif self.position_modifier == SequenceLocation.c_term:
            parts.append("C-term")
        elif self.position_modifier == SequenceLocation.protein_c_term:
            parts.append("Protein C-term")
        return " ".join(parts)



from glypy.composition.ccomposition cimport CComposition

from glycopeptidepy._c.structure.fragment cimport IonSeriesBase
from glycopeptidepy._c.structure.sequence_methods cimport _PeptideSequenceCore
from glycopeptidepy._c.structure.base cimport ModificationBase

from glycopeptidepy.structure.modification import (
    Modification,
    NGlycanCoreGlycosylation,
    OGlycanCoreGlycosylation,
    GlycosaminoglycanLinkerGlycosylation)


from glycopeptidepy.structure.glycan import (GlycosylationType)


cdef:
    ModificationBase _n_glycosylation
    ModificationBase _o_glycosylation
    ModificationBase _gag_linker_glycosylation
    ModificationBase _modification_hexnac
    ModificationBase _modification_xylose

    dict glycosylation_type_to_core



cdef class FragmentationStrategyBase(object):
    cdef:
        public _PeptideSequenceCore peptide
        public bint compute_compositions

    cpdef CComposition peptide_composition(self)
    cpdef CComposition total_composition(self)




try:
    from glycopeptidepy._c.count_table import CountTable

    class ModificationIndex(CountTable):
        pass
except ImportError:
    class ModificationIndex(dict):
        def __getitem__(self, key):
            try:
                return dict.__getitem__(self, key)
            except KeyError:
                return 0


class ModificationStringParseError(ValueError):
    pass


class ModificationNameResolutionError(KeyError):
    pass



from cpython.object cimport PyObject
from cpython.dict cimport PyDict_GetItem
from glycopeptidepy._c.structure.modification.rule cimport ModificationRuleBase

from glycopeptidepy.structure.modification.target import title_cleaner
from glycopeptidepy.structure.modification.utils import ModificationStringParseError
from glycopeptidepy.structure.modification.rule import ModificationRule


cdef title_cleaner_search = title_cleaner.search
cdef object ModificationRuleImpl = ModificationRule


cdef basestring clean_name(basestring key):
    match = title_cleaner_search(key)
    if match is not None:
        return match.groupdict()["name"]
    raise ModificationStringParseError("Could not parse {}".format(key))


cdef class ModificationTableBase(object):
    '''A mixin class to provide implementation for :class:`ModificationSource`
    methods for :class:`ModificationTable`
    '''

    def __getitem__(self, basestring key):
        return self.resolve(key)

    cpdef ModificationRuleBase resolve(self, basestring key):
        try:
            rule = self._find_modification_rule(key)
        except ModificationStringParseError:
            rule = ModificationRuleImpl.resolve(key)
        return rule

    cpdef get_modification(self, name):
        return self.resolve(name)()

    cdef ModificationRuleBase _find_modification_rule(self, basestring key):
        cdef:
            PyObject* ptemp

        # First try the key in the custom mapping
        ptemp = PyDict_GetItem(self._instance_custom_rules, key)
        if ptemp != NULL:
            return <ModificationRuleBase>ptemp
        # Next try the normal store
        ptemp = PyDict_GetItem(self.store, key)
        if ptemp != NULL:
            return <ModificationRuleBase>ptemp
        # Otherwise try to look fuzzy
        return self._find_modification_rule_fuzzy(key)

    cdef ModificationRuleBase _find_modification_rule_fuzzy(self, basestring key):
        cdef:
            PyObject* ptemp
            basestring name
        # Clean the key in case it is decorated with extra target information
        # or throw an error
        name = clean_name(key)
        ptemp = PyDict_GetItem(self.store, name)
        if ptemp != NULL:
            return <ModificationRuleBase>ptemp
        ptemp = PyDict_GetItem(self._instance_custom_rules, name)
        if ptemp != NULL:
            return <ModificationRuleBase>ptemp
        raise ModificationStringParseError("Could not resolve {0}".format(name))


import unittest

from glycopeptidepy.structure import residue


class AminoAcidResidueTests(unittest.TestCase):
    def test_residue_creation(self):
        leucine = residue.AminoAcidResidue(symbol='L')
        isoleucine = residue.AminoAcidResidue(name='Ile')
        self.assertEqual(leucine.composition, isoleucine.composition)

    def test_degeneracy(self):
        xle = residue.AminoAcidResidue("J")
        leucine = residue.AminoAcidResidue(symbol='L')
        self.assertTrue(xle.is_degenerate)
        self.assertFalse(leucine.is_degenerate)
        self.assertTrue(leucine in residue.degeneracy_index[xle.name])

        self.assertNotIn(leucine, residue.get_all_sequencing_residues())



# pylint: disable=unsubscriptable-object

from collections import defaultdict
from collections.abc import Mapping

from glypy import Composition
from glypy.utils import Enum
from glypy.structure.glycan import NamedGlycan
from glypy.structure.glycan_composition import (
    FrozenMonosaccharideResidue,
    GlycanComposition,
    HashableGlycanComposition)

from glycopeptidepy.utils.collectiontools import decoratordict


class allset(object):
    def __contains__(self, x):
        return True


class GlycosylationType(Enum):
    n_linked = 1
    o_linked = 2
    glycosaminoglycan = 3
    other = 4


GlycosylationType.n_linked.add_name("N-Linked")
GlycosylationType.n_linked.add_name("N-linked")
GlycosylationType.n_linked.add_name("n-linked")
GlycosylationType.n_linked.add_name("N-Glycan")
GlycosylationType.n_linked.add_name("n-glycan")
GlycosylationType.n_linked.add_name("n_glycan")
GlycosylationType.n_linked.name = "N-Linked"

GlycosylationType.o_linked.add_name("O-Linked")
GlycosylationType.o_linked.add_name("O-linked")
GlycosylationType.o_linked.add_name("o-linked")
GlycosylationType.o_linked.add_name("O-Glycan")
GlycosylationType.o_linked.add_name("o-glycan")
GlycosylationType.o_linked.add_name("o_glycan")
GlycosylationType.o_linked.name = "O-Linked"

GlycosylationType.glycosaminoglycan.add_name("Glycosaminoglycan")
GlycosylationType.glycosaminoglycan.add_name("GAG-Linker")
GlycosylationType.glycosaminoglycan.add_name("gag-linker")
GlycosylationType.glycosaminoglycan.add_name("gag_linker")
GlycosylationType.glycosaminoglycan.name = "Glycosaminoglycan"


glycosylation_site_detectors = decoratordict({
    GlycosylationType.n_linked: lambda x: [],
    GlycosylationType.o_linked: lambda x: [],
    GlycosylationType.glycosaminoglycan: lambda x: []
})


class TypedGlycanComposition(HashableGlycanComposition):
    glycosylation_type: GlycosylationType

    def __init__(self, glycosylation_type=None, *args, **kwargs):
        if isinstance(glycosylation_type, GlycanComposition):
            temp = args[0]
            args = (glycosylation_type,) + args[1:]
            glycosylation_type = temp
        self.glycosylation_type = GlycosylationType[glycosylation_type]
        super(TypedGlycanComposition, self).__init__(*args, **kwargs)

    @classmethod
    def _empty(cls):
        inst = super(cls, cls)._empty()
        inst.glycosylation_type = None
        return inst

    def clone(self, propogate_composition_offset=True):
        dup = self.__class__(self.glycosylation_type, self)
        if not propogate_composition_offset:
            dup.composition_offset = Composition('H2O')
        return dup

    def is_type(self, glycosylation_type):
        return self.glycosylation_type is GlycosylationType[glycosylation_type]

    def as_modification(self):
        from .modification import Glycosylation
        return Glycosylation(self, metadata={
            "glycosylation_type": self.glycosylation_type.name
        })


class GlycanCombination(Mapping):  # pragma: no cover
    def __init__(self, components):
        self.components = tuple(components)
        self._glycan_composition = self._sum_components(self.components)

    def _sum_components(self, components):
        acc = HashableGlycanComposition()
        for x in components:
            acc += x
        acc.composition_offset -= Composition({"H": 2, "O": 1})
        return acc

    def mass(self, *args, **kwargs):
        return self._glycan_composition.mass(*args, **kwargs)

    def total_composition(self, *args, **kwargs):
        return self._glycan_composition.total_composition(*args, **kwargs)

    def __iter__(self):
        return iter(self._glycan_composition)

    def __getitem__(self, key):
        return self._glycan_composition[key]

    def keys(self):
        return self._glycan_composition.keys()

    def values(self):
        return self._glycan_composition.values()

    def items(self):
        return self._glycan_composition.items()

    def clone(self, *args, **kwargs):
        return self._glycan_composition.clone(*args, **kwargs)

    def serialize(self):
        parts = [str(x)[1:-1] for x in self.components]
        return '{%s}' % '|'.join(parts)

    def __repr__(self):
        return self.serialize()

    def __str__(self):
        return self.serialize()

    @classmethod
    def parse(cls, string):
        string = str(string)[1:-1]
        parts = string.split("|")
        components = []
        for part in parts:
            component = HashableGlycanComposition()
            tokens = part.split("; ")
            for token in tokens:
                try:
                    residue, count = token.split(":")
                except ValueError:
                    if string == "{}":
                        return cls([])
                    else:
                        raise ValueError("Malformed Token, %s" % (token,)) from None
                component[FrozenMonosaccharideResidue.from_iupac_lite(residue)] = int(count)
            components.append(component)
        return cls(components)

    def __hash__(self):
        return hash(self._glycan_composition)

    def __eq__(self, other):
        try:
            other = other._glycan_composition
        except AttributeError:
            pass
        return self._glycan_composition == other

    def __ne__(self, other):
        try:
            other = other._glycan_composition
        except AttributeError:
            pass
        return (self._glycan_composition != other)

    def __add__(self, other):
        return self._glycan_composition + other

    def __sub__(self, other):
        return self._glycan_composition - other

    def __mul__(self, other):
        return self._glycan_composition * other

    def __len__(self):
        return len(self._glycan_composition)


class TypedGlycan(NamedGlycan):
    glycosylation_type: GlycosylationType

    def __init__(self, glycosylation_type=None, *args, **kwargs):
        self.glycosylation_type = GlycosylationType[glycosylation_type]
        super(TypedGlycan, self).__init__(*args, **kwargs)

    def clone(self, index_method='dfs', cls=None):
        if cls is None:
            cls = TypedGlycan
        inst = super(TypedGlycan, self).clone(index_method=index_method, cls=cls)
        inst.glycosylation_type = self.glycosylation_type
        return inst

    def __repr__(self):
        rep = super(TypedGlycan, self).__repr__()
        return "%s\n%s" % (self.glycosylation_type, rep)

    def is_type(self, glycosylation_type):
        return self.glycosylation_type == GlycosylationType[glycosylation_type]

    def as_modification(self, encoding_format=None):
        from .modification import Glycosylation
        return Glycosylation(self, encoding_format, metadata={
            "glycosylation_type": self.glycosylation_type.name
        })


class GlycanCompositionProxy(Mapping):
    '''A mapping-like object that imitates the GlycanComposition interface in
    a read-only fashion.
    '''

    def __init__(self, glycan_composition):
        self.obj = glycan_composition

    def mass(self, *args, **kwargs):
        return self.obj.mass(*args, **kwargs)

    def total_composition(self, *args, **kwargs):
        return self.obj.total_composition(*args, **kwargs)

    def __iter__(self):
        return iter(self.obj)

    def __getitem__(self, key):
        return self.obj[key]

    def _getitem_fast(self, key):
        return self.obj._getitem_fast(key)

    def __contains__(self, key):
        return key in self.obj

    def keys(self):
        return self.obj.keys()

    def values(self):
        return self.obj.values()

    def items(self):
        return self.obj.items()

    def clone(self, *args, **kwargs):
        return self.obj.clone(*args, **kwargs)

    def query(self, query, exact=True, **kwargs):
        return self.obj.query(query, exact=exact, **kwargs)

    def __repr__(self):
        return repr(self.obj)

    def __str__(self):
        return str(self.obj)

    def __hash__(self):
        return hash(self.obj)

    def __eq__(self, other):
        try:
            other = other.obj
        except AttributeError:
            pass
        return self.obj == other

    def __ne__(self, other):
        try:
            other = other.obj
        except AttributeError:
            pass
        return (self.obj != other)

    def __add__(self, other):
        return self.obj + other

    def __sub__(self, other):
        return self.obj - other

    def __mul__(self, other):
        return self.obj * other

    def __len__(self):
        return len(self.obj)


try:
    from glycopeptidepy._c.structure.glycan import GlycanCompositionProxy
except ImportError:
    pass


class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):
    """A :class:`GlycanCompositionProxy` that pretends to allow you to
    mutate the :attr:`composition_offset` attribute of the underlying
    composition, but instead stores that within the proxy, leaving the
    underlying compostion unchanged.

    This type should be useful for re-using the same :class:`FrozenGlycanComposition`
    object without repeatedly copying and invalidating its cached properties. When
    this object is copied, the underlying GlycanComposition is not copied, only
    the proxy.
    """
    def __init__(self, obj, offset=None):
        if offset is None:
            offset = Composition()
        super().__init__(obj)
        self.composition_offset = offset

    def clone(self, *args, **kwargs):
        return self.__class__(self.obj, self.composition_offset.copy())

    def mass(self, *args, **kwargs):
        mass = self.obj.mass(*args, **kwargs)
        return mass + self.composition_offset.mass

    def total_composition(self, *args, **kwargs):
        comp = self.obj.total_composition(*args, **kwargs)
        return comp + self.composition_offset


try:
    from glycopeptidepy._c.structure.glycan import GlycanCompositionWithOffsetProxy
except ImportError:
    pass


WATER_OFFSET = Composition({"H": 2, "O": 1})

class GlycosylationManager(object):
    def __init__(self, parent, aggregate=None):
        self.mapping = {}
        self.parent = parent
        self._aggregate = None
        self._proxy = None
        self._type_track = None
        if aggregate is not None:
            self.set_aggregate(aggregate)
        self._total_glycosylation_size = None

    def clear(self):
        self.invalidate()
        self.aggregate = None
        self.mapping.clear()

    def __setitem__(self, key, value):
        self.invalidate()
        self.mapping[key] = value

    def __delitem__(self, key):
        self.invalidate()
        del self.mapping[key]

    def pop(self, key, default=None):
        self.invalidate()
        self.mapping.pop(key, default)

    def update(self, base):
        self.invalidate()
        self.mapping.update(base)

    def keys(self):
        return self.mapping.keys()

    def values(self):
        return self.mapping.values()

    def items(self):
        return self.mapping.items()

    def __getitem__(self, key):
        return self.mapping[key]

    def __len__(self):
        return len(self.mapping)

    def invalidate(self):
        self._proxy = None
        self._type_track = None
        self._total_glycosylation_size = None

    @property
    def glycosylation_types(self):
        if self._type_track is None:
            self._type_track = self._update_track()
        return self._type_track

    def count_glycosylation_type(self, glycosylation_type):
        return len(self.glycosylation_types[glycosylation_type])

    def _update_track(self):
        track = defaultdict(list)
        for position, value in self.items():
            glycosylation = value.rule
            track[glycosylation.glycosylation_type].append((position, value))
        return track

    setdefault = None
    popitem = None

    def copy(self):
        inst = self.__class__(
            self.parent,
            self.aggregate.clone() if self.aggregate is not None else None)
        inst.update(self)
        return inst

    @property
    def aggregate(self):
        return self._aggregate

    @aggregate.setter
    def aggregate(self, value):
        self._aggregate = value
        if self._aggregate is not None:
            self._patch_aggregate()
        self.invalidate()

    def _patch_aggregate(self):
        self.aggregate.composition_offset -= WATER_OFFSET

    @property
    def glycan_composition(self):
        if self._proxy is None:
            self._proxy = self._make_glycan_composition_proxy()
        return self._proxy

    def total_glycosylation_size(self):
        if self._total_glycosylation_size is None:
            self._total_glycosylation_size = sum(self.aggregate.values())
        return self._total_glycosylation_size

    def clone(self):
        return self.copy()

    def total_composition(self):
        total = Composition()
        has_aggregate = self.aggregate is not None
        for _key, value in self.items():
            if has_aggregate and value.rule.is_core:
                continue
            total += value.composition
        if has_aggregate:
            total += self.aggregate.total_composition()
        return total

    def mass(self):
        total = 0.
        has_aggregate = self.aggregate is not None
        for _key, value in self.items():
            if has_aggregate and value.rule.is_core:
                continue
            total += value.mass
        if has_aggregate:
            total += self.aggregate.mass()
        return total

    def is_fully_specified_topologies(self):
        is_fully_specified = len(self) > 0
        for _key, value in self.items():
            if value.rule.is_core:
                is_fully_specified = False
                break
            elif value.rule.is_composition:
                is_fully_specified = False
                break
        return is_fully_specified

    def _make_glycan_composition_proxy(self):
        if self.aggregate is not None:
            base = self.aggregate.clone()
        else:
            base = HashableGlycanComposition()
            # Represent the initial amide bond between the peptide
            # and the first glycan. Subsequent glycans do not need
            # further chemical losses because of the dehyration built
            # directly into the Residue abstraction.
            base.composition_offset -= WATER_OFFSET
        for _key, value in self.items():
            if value.rule.is_core:
                continue
            elif value.rule.is_composition:
                base += value.rule.glycan
            else:
                # Convert Glycan object into a composition, using the original
                # detatched topology to omit the "aglycone" group which represents
                # the connection between the glycan and the peptide, which penalizes
                # the composition by H2O. This H2O is lost when that bond is formed,
                # but doesn't need to be explicitly included as the loss is tracked
                # when initializing the base above.
                gc = HashableGlycanComposition.from_glycan(
                    value.rule._original)
                base += gc
        return GlycanCompositionProxy(base)


try:
    from glycopeptidepy._c.structure.glycan import GlycosylationManager
except ImportError:
    pass


Mapping.register(GlycanCompositionProxy)
Mapping.register(GlycosylationManager)
